from decimal import Decimal
from autotrader.brokers.broker_utils import BrokerUtils


class Utils(BrokerUtils):
    def __init__(self, **kwargs):
        # Create Client instance
        try:
            from dydx3 import Client

            self.api = Client(host="https://api.dydx.exchange")
        except ImportError:
            raise Exception(
                "Please install dydx-v3-python to connect " + "to dydx API."
            )
        # Except network error?

        # Stored instrument precisions
        self._instrument_precisions = {}

    def __repr__(self):
        return "AutoTrader-dYdX Broker Utilities"

    def __str__(self):
        return "AutoTrader-dYdX Broker Utilities"

    def get_precision(self, instrument, *arg, **kwargs):
        """Returns the precision of the specified instrument."""
        if instrument in self._instrument_precisions:
            # Precision already fetched, use stored value
            unified_response = self._instrument_precisions[instrument]
        else:
            # Fetch precision
            market = self._get_market(instrument=instrument)
            unified_response = {
                "size": -Decimal(market["stepSize"]).as_tuple().exponent,
                "price": -Decimal(market["tickSize"]).as_tuple().exponent,
            }

            # Store for later use
            self._instrument_precisions[instrument] = unified_response

        return unified_response

    def _get_market(self, instrument, *args, **kwargs):
        """Returns the raw get_market response from dYdX"""
        response = self.api.public.get_markets(instrument)
        return response.data["markets"][instrument]



import os
import telegram
from autotrader.brokers.trading import Order
from autotrader.comms.notifier import Notifier
from autotrader.utilities import read_yaml, write_yaml
from telegram.ext import Updater, CommandHandler, MessageHandler, Filters


class Telegram(Notifier):
    def __init__(self, api_token: str = None, chat_id: str = None) -> None:
        self.api_token = api_token
        self.chat_id = chat_id

        if api_token is not None:
            self.bot = telegram.Bot(api_token)

    def __repr__(self) -> str:
        return "AutoTrader-Telegram communication module"

    def send_order(self, order: Order, *args, **kwargs) -> None:
        side = "long" if order.direction > 0 else "short"
        message = (
            f"New {order.instrument} {order.order_type} order created: "
            + f"{order.size} units {side}"
        )

        # Create bot and send message
        self.bot.send_message(chat_id=self.chat_id, text=message)

    def send_message(self, message: str, *args, **kwargs) -> None:
        """A generic method to send a custom message.

        Parameters
        ----------
        message : str
            The message to be sent.
        """
        # Send message
        self.bot.send_message(chat_id=self.chat_id, text=message)

    def run_bot(self) -> None:
        """A method to initialise your bot and get your chat ID.
        This method should be running before you send any messages
        to it on Telegram. To start, message the BotFather on
        Telegram, and create a new bot. Use the API token provided
        to run this method.

        Parameters
        ----------
        api_token : str
            The API token of your telegram bot.

        Returns
        -------
        None.

        """
        # Check for api token
        if self.api_token is None:
            path = "config/keys.yaml"
            if os.path.exists(path):
                # Look to fetch api token from config file
                config = read_yaml(path)
                if "TELEGRAM" in config:
                    if (
                        config["TELEGRAM"] is not None
                        and "api_key" in config["TELEGRAM"]
                    ):
                        self.api_token = config["TELEGRAM"]["api_key"]

                    else:
                        print(
                            "Please add your Telegram API key to the "
                            + "keys.yaml file."
                        )
                        return
                else:
                    print(
                        "Please add a 'TELEGRAM' key to the keys.yaml "
                        + "file, with a sub-key for the 'api_key'."
                    )
                    return

        updater = Updater(self.api_token, use_context=True)
        dp = updater.dispatcher

        dp.add_handler(CommandHandler("start", self._start_command))
        dp.add_handler(CommandHandler("help", self._help_command))

        dp.add_handler(MessageHandler(Filters.text, self._handle_message))

        updater.start_polling()
        updater.idle()

    @staticmethod
    def _start_command(
        update,
        context,
    ):
        # Extract user name and chat ID
        name = update.message.chat.first_name
        chat_id = str(update.message.chat_id)

        # Create response
        response = (
            f"Hi {name}, welcome to your very own AutoTrader "
            + f"Telegram bot. Your chat ID is {chat_id}. Use this to "
            + "set-up trading notifications. Note that this ID has also "
            + "been printed to your computer screen for reference."
        )
        print(f"Chat ID: {chat_id}")

        # Send response
        update.message.reply_text(response)

    @staticmethod
    def _help_command(
        update,
        context,
    ):
        update.message.reply_text("Help is on the way!")

    # @staticmethod
    def _handle_message(self, update, context):
        text = str(update.message.text).lower()

        if "id" in text:
            chat_id = str(update.message.chat_id)

            if "write" in text:
                # Write chat ID to keys.yaml file
                path = "config/keys.yaml"
                if os.path.exists(path):
                    # Config file exists, proceed
                    config = read_yaml(path)

                    if "TELEGRAM" in config:
                        config["TELEGRAM"]["chat_id"] = chat_id
                    else:
                        config["TELEGRAM"] = {
                            "api_key": self.api_token,
                            "chat_id": chat_id,
                        }

                    # Write to file
                    write_yaml(config, path)

                    response = "All done."

                else:
                    response = (
                        "I couldn't find your keys.yaml directory. "
                        + "Make sure you are running the bot from your project "
                        + "home directory, with config/keys.yaml within it."
                    )

            else:
                # Return chat ID
                response = f"Your chat ID is {chat_id}."

                if "print" in text or "show" in text:
                    print(f"Chat ID: {chat_id}")
                    response += " This has also been printed to your computer."

        elif "thank" in text or "ty" in text:
            response = "You're welcome."

        else:
            response = "I'm not quite ready to respond to that..."
        update.message.reply_text(response)

    @staticmethod
    def _error(update, context):
        print(f"Update {update} caused error {context.error}")



from autotrader.brokers.trading import Order


class Strategy:
    def __init__(self, instrument, parameters, **kwargs):
        """Initialise the strategy."""
        self.name = "Template Strategy"
        self.instrument = instrument
        self.params = parameters

        # Construct indicators dict for plotting
        self.indicators = {
            "Indicator Name": {"type": "indicatortype", "data": "indicatordata"},
        }

    def generate_signal(self, data):
        """Define strategy logic to transform data into trading signals."""

        # Initialise empty order list
        orders = []

        # The data passed into this method is the most up-to-date data.

        # Example long market order:
        long_market_order = Order(direction=1)
        orders.append(long_market_order)

        # Example short limit order:
        short_limit = Order(direction=-1, order_type="limit", order_limit_price=1.0221)
        orders.append(short_limit)

        # Return any orders generated
        # If no orders are generated, return an empty list [], an empty dict {},
        # or a blank order Order().
        return orders



import os
import importlib
import traceback
import pandas as pd
from autotrader.comms import emailing
from datetime import datetime, timezone
from autotrader.autodata import AutoData
from autotrader.brokers.trading import Order
from concurrent.futures import ThreadPoolExecutor
from autotrader.utilities import get_data_config, TradeAnalysis


class AutoTraderBot:
    """AutoTrader Trading Bot.

    Attributes
    ----------
    instrument : str
        The trading instrument assigned to the bot.
    data : pd.DataFrame
        The OHLC price data used by the bot.
    quote_data : pd.DataFrame
        The OHLC quote data used by the bot.
    MTF_data : dict
        The multiple timeframe data used by the bot.
    backtest_results : TradeAnalysis
        A class containing results from the bot in backtest. This
        is available only after a backtest.
    """

    def __init__(
        self,
        instrument: str,
        strategy_dict: dict,
        broker,
        deploy_dt: datetime,
        data_dict: dict,
        quote_data_path: str,
        auxdata: dict,
        autotrader_instance,
    ) -> None:
        """Instantiates an AutoTrader Bot.

        Parameters
        ----------
        instrument : str
            The trading instrument assigned to the bot instance.
        strategy_dict : dict
            The strategy configuration dictionary.
        broker : AutoTrader Broker instance
            The AutoTrader Broker module.
        deploy_dt : datetime
            The datetime stamp of the bot deployment time.
        data_dict : dict
            The strategy data.
        quote_data_path : str
            The quote data filepath for the trading instrument
            (for backtesting only).
        auxdata : dict
            Auxiliary strategy data.
        autotrader_instance : AutoTrader
            The parent AutoTrader instance.

        Raises
        ------
        Exception
            When there is an error retrieving the instrument data.

        Returns
        -------
        None
            The trading bot will be instantiated and ready for trading.

        """
        # Inherit user options from autotrader
        for attribute, value in autotrader_instance.__dict__.items():
            setattr(self, attribute, value)

        # Assign local attributes
        self.instrument = instrument
        self._broker = broker

        # # Define execution framework
        if self._execution_method is None:
            self._execution_method = self._submit_order

        # Check for muliple brokers and construct mapper
        if self._multiple_brokers:
            # Trading across multiple venues
            self._brokers = self._broker
            self._instrument_to_broker = {}
            for (
                broker_name,
                tradeable_instruments,
            ) in self._virtual_tradeable_instruments.items():
                for instrument in tradeable_instruments:
                    if instrument in self._instrument_to_broker:
                        # Instrument is already in mapper, add broker
                        self._instrument_to_broker[instrument].append(
                            self._brokers[broker_name]
                        )
                    else:
                        # New instrument, add broker
                        self._instrument_to_broker[instrument] = [
                            self._brokers[broker_name]
                        ]

        else:
            # Trading through a single broker
            self._brokers = {self._broker_name: self._broker}

            # Map instruments to broker
            self._instrument_to_broker = {}
            instruments = [instrument] if isinstance(instrument, str) else instrument
            for instrument in instruments:
                self._instrument_to_broker[instrument] = [self._broker]

        # Unpack strategy parameters and assign to strategy_params
        strategy_config = strategy_dict["config"]
        interval = strategy_config["INTERVAL"]
        period = strategy_config["PERIOD"]
        risk_pc = strategy_config["RISK_PC"] if "RISK_PC" in strategy_config else None
        sizing = strategy_config["SIZING"] if "SIZING" in strategy_config else None
        params = (
            strategy_config["PARAMETERS"] if "PARAMETERS" in strategy_config else {}
        )
        strategy_params = params
        strategy_params["granularity"] = (
            strategy_params["granularity"]
            if "granularity" in strategy_params
            else interval
        )
        strategy_params["risk_pc"] = (
            strategy_params["risk_pc"] if "risk_pc" in strategy_params else risk_pc
        )
        strategy_params["sizing"] = (
            strategy_params["sizing"] if "sizing" in strategy_params else sizing
        )
        strategy_params["period"] = (
            strategy_params["period"] if "period" in strategy_params else period
        )
        strategy_params["INCLUDE_POSITIONS"] = (
            strategy_config["INCLUDE_POSITIONS"]
            if "INCLUDE_POSITIONS" in strategy_config
            else False
        )
        strategy_config["INCLUDE_BROKER"] = (
            strategy_config["INCLUDE_BROKER"]
            if "INCLUDE_BROKER" in strategy_config
            else False
        )
        strategy_config["INCLUDE_STREAM"] = (
            strategy_config["INCLUDE_STREAM"]
            if "INCLUDE_STREAM" in strategy_config
            else False
        )
        self._strategy_params = strategy_params

        # Import Strategy
        if strategy_dict["class"] is not None:
            strategy = strategy_dict["class"]
        else:
            strat_module = strategy_config["MODULE"]
            strat_name = strategy_config["CLASS"]
            strat_package_path = os.path.join(self._home_dir, "strategies")
            strat_module_path = os.path.join(strat_package_path, strat_module) + ".py"
            strat_spec = importlib.util.spec_from_file_location(
                strat_module, strat_module_path
            )
            strategy_module = importlib.util.module_from_spec(strat_spec)
            strat_spec.loader.exec_module(strategy_module)
            strategy = getattr(strategy_module, strat_name)

        # Strategy shutdown routine
        self._strategy_shutdown_method = strategy_dict["shutdown_method"]

        # Get data feed configuration
        data_config = get_data_config(
            feed=self._feed,
            global_config=self._global_config_dict,
            environment=self._environment,
        )

        # Data retrieval
        self._quote_data_file = quote_data_path  # Either str or None
        self._data_filepaths = data_dict  # Either str or dict, or None
        self._auxdata_files = auxdata  # Either str or dict, or None

        if self._feed == "none":
            # None data-feed being used, allow duplicate bars
            self._allow_duplicate_bars = True

        # Check for portfolio strategy
        trade_portfolio = (
            strategy_config["PORTFOLIO"] if "PORTFOLIO" in strategy_config else False
        )

        portfolio = strategy_config["WATCHLIST"] if trade_portfolio else False

        # Fetch data
        self._get_data = AutoData(
            data_config, self._allow_dancing_bears, self._base_currency
        )

        # Create instance of data stream object
        stream_attributes = {
            "data_filepaths": self._data_filepaths,
            "quote_data_file": self._quote_data_file,
            "auxdata_files": self._auxdata_files,
            "strategy_params": self._strategy_params,
            "get_data": self._get_data,
            "data_start": self._data_start,
            "data_end": self._data_end,
            "instrument": self.instrument,
            "feed": self._feed,
            "portfolio": portfolio,
            "data_path_mapper": self._data_path_mapper,
            "data_dir": self._data_directory,
            "backtest_mode": self._backtest_mode,
        }
        self.Stream = self._data_stream_object(**stream_attributes)

        # Initial data call
        self._refresh_data(deploy_dt)

        # Instantiate Strategy
        strategy_inputs = {
            "parameters": params,
            "data": self._strat_data,
            "instrument": self.instrument,
        }

        if strategy_config["INCLUDE_BROKER"]:
            strategy_inputs["broker"] = self._broker
            strategy_inputs["broker_utils"] = self._broker_utils

        if strategy_config["INCLUDE_STREAM"]:
            strategy_inputs["data_stream"] = self.Stream

        my_strat = strategy(**strategy_inputs)

        # Assign strategy to local attributes
        self._last_bars = None
        self._strategy = my_strat
        self._strategy_name = (
            strategy_config["NAME"]
            if "NAME" in strategy_config
            else "(unnamed strategy)"
        )

        # Assign strategy attributes for tick-based strategy development
        if self._backtest_mode:
            self._strategy._backtesting = True
            self.trade_results = None
        if interval.split(",")[0] == "tick":
            self._strategy._tick_data = True

    def __repr__(self):
        # TODO - alter str for portfolio bots
        if isinstance(self.instrument, list):
            return "Portfolio AutoTraderBot"
        else:
            return f"{self.instrument} AutoTraderBot"

    def __str__(self):
        return "AutoTraderBot instance"

    def _update(self, i: int = None, timestamp: datetime = None) -> None:
        """Update strategy with the latest data and generate a trade signal.

        Parameters
        ----------
        i : int, optional
            The indexing parameter used when running in periodic update mode.
            The default is None.
        timestamp : datetime, optional
            The timestamp parameter used when running in continuous update
            mode. The default is None.

        Returns
        -------
        None
            Trade signals generated will be submitted to the broker.
        """

        if self._run_mode == "continuous":
            # Running in continuous update mode
            strat_data, current_bars, quote_bars, sufficient_data = self._check_data(
                timestamp, self._data_indexing
            )
            strat_object = strat_data

        else:
            # Running in periodic update mode
            current_bars = {self.instrument: self.data.iloc[i]}
            quote_bars = {self.instrument: self.quote_data.iloc[i]}
            sufficient_data = True
            strat_object = i

        # Check for new data
        new_data = self._check_last_bar(current_bars)

        if sufficient_data and new_data:
            # There is a sufficient amount of data, and it includes new data
            if self._backtest_mode or self._papertrading:
                # Update virtual broker with latest price bars
                self._update_virtual_broker(current_bars)

            # Get strategy orders
            strategy_orders = self._strategy.generate_signal(strat_object)

            # Check and qualify orders
            orders = self._check_orders(strategy_orders)
            self._qualify_orders(orders, current_bars, quote_bars)

            if not self._scan_mode:
                # Submit orders
                if self._max_workers is not None:
                    workers = min(self._max_workers, len(orders))
                else:
                    workers = None
                with ThreadPoolExecutor(max_workers=workers) as executor:
                    futures = []
                    for order in orders:
                        try:
                            order_time = current_bars[order.instrument].name
                        except:
                            if self._feed == "none":
                                order_time = datetime.now(timezone.utc)
                            else:
                                order_time = current_bars[order.data_name].name

                        # Submit order to relevant exchange
                        futures.append(
                            executor.submit(
                                self._execution_method,
                                broker=self._brokers[order.exchange],
                                order=order,
                                order_time=order_time,
                            )
                        )

                # Check for exceptions
                for f in futures:
                    try:
                        f.result()
                    except Exception as e:
                        traceback_str = "".join(traceback.format_tb(e.__traceback__))
                        exception_str = (
                            f"AutoTrader exception when submitting order: {e}"
                        )
                        print_str = exception_str + "\nTraceback:\n" + traceback_str
                        print(print_str)

            if self._papertrading:
                # Update virtual broker again to trigger any orders
                self._update_virtual_broker(current_bars)

            if int(self._verbosity) > 1:
                try:
                    current_time = current_bars[
                        list(current_bars.keys())[0]
                    ].name.strftime("%b %d %Y %H:%M:%S")
                except:
                    current_time = datetime.now().strftime("%b %d %Y %H:%M:%S")
                if len(orders) > 0:
                    for order in orders:
                        direction = "long" if order.direction > 0 else "short"
                        order_string = (
                            f"{current_time}: {order.instrument} "
                            + f"{direction} {order.order_type} order of "
                            + f"{order.size} units placed."
                        )
                        print(order_string)
                else:
                    if int(self._verbosity) > 2:
                        print(
                            f"{current_time}: No signal detected ({self.instrument})."
                        )

            # Check for orders placed and/or scan hits
            if int(self._notify) > 0 and not (self._backtest_mode or self._scan_mode):
                for order in orders:
                    self._notifier.send_order(order)

            # Check scan results
            if self._scan_mode:
                # Report AutoScan results
                if int(self._verbosity) > 0 or int(self._notify) == 0:
                    # Scan reporting with no notifications requested
                    if len(orders) == 0:
                        print("{}: No signal detected.".format(self.instrument))

                    else:
                        # Scan detected hits
                        print("Scan hits:")
                        for order in orders:
                            print(order)

                if int(self._notify) > 0:
                    # Notifications requested
                    for order in orders:
                        self._notifier.send_message(f"Scan hit: {order}")

        else:
            if int(self._verbosity) > 1:
                print(
                    "\nThe strategy has not been updated as there is either "
                    + "insufficient data, or no new data. If you believe "
                    + "this is an error, try setting allow_dancing_bears to "
                    + "True, or set allow_duplicate_bars to True in "
                    + "AutoTrader.configure().\n"
                    + f"Sufficient data: {sufficient_data}\n"
                    + f"New data: {new_data}"
                )

    def _refresh_data(self, timestamp: datetime = None, **kwargs):
        """Refreshes the active Bot's data attributes for trading.

        When backtesting without dynamic data updates, the data attributes
        of the bot will be constant. When using dynamic data, or when
        livetrading in continuous mode, the data attributes will change
        as time passes, reflecting more up-to-date data. This method refreshes
        the data attributes for a given timestamp by calling the datastream
        object.

        Parameters
        ----------
        timestamp : datetime, optional
            The current timestamp. If None, datetime.now() will be called.
            The default is None.
        **kwargs : dict
            Any other named arguments.

        Raises
        ------
        Exception
            When there is an error retrieving the data.

        Returns
        -------
        None:
            The up-to-date data will be assigned to the Bot instance.

        """

        timestamp = datetime.now(timezone.utc) if timestamp is None else timestamp

        # Fetch new data
        data, multi_data, quote_data, auxdata = self.Stream.refresh(timestamp=timestamp)

        # Check data returned is valid
        if self._feed != "none" and len(data) == 0:
            raise Exception("Error retrieving data.")

        # Data assignment
        if multi_data is None:
            strat_data = data
        else:
            strat_data = multi_data

        # Auxiliary data assignment
        if auxdata is not None:
            strat_data = {"base": strat_data, "aux": auxdata}

        # Assign data attributes to bot
        self._strat_data = strat_data
        self.data = data
        self.multi_data = multi_data
        self.auxdata = auxdata
        self.quote_data = quote_data

    def _check_orders(self, orders) -> list:
        """Checks that orders returned from strategy are in the correct
        format.

        Returns
        -------
        List of Orders

        Notes
        -----
        An order must have (at the very least) an order type specified. Usually,
        the direction will also be required, except in the case of close order
        types. If an order with no order type is provided, it will be ignored.
        """

        def check_type(orders):
            checked_orders = []
            if isinstance(orders, dict):
                # Order(s) provided in dictionary
                if "order_type" in orders:
                    # Single order dict provided
                    if "instrument" not in orders:
                        orders["instrument"] = self.instrument
                    checked_orders.append(Order._from_dict(orders))

                elif len(orders) > 0:
                    # Multiple orders provided
                    for key, item in orders.items():
                        if isinstance(item, dict) and "order_type" in item:
                            # Convert order dict to Order object
                            if "instrument" not in item:
                                item["instrument"] = self.instrument
                            checked_orders.append(Order._from_dict(item))
                        elif isinstance(item, Order):
                            # Native Order object, append as is
                            checked_orders.append(item)
                        else:
                            raise Exception(f"Invalid order submitted: {item}")

                elif len(orders) == 0:
                    # Empty order dict
                    pass

            elif isinstance(orders, Order):
                # Order object directly returned
                checked_orders.append(orders)

            elif isinstance(orders, list):
                # Order(s) provided in list
                for item in orders:
                    if isinstance(item, dict) and "order_type" in item:
                        # Convert order dict to Order object
                        if "instrument" not in item:
                            item["instrument"] = self.instrument
                        checked_orders.append(Order._from_dict(item))
                    elif isinstance(item, Order):
                        # Native Order object, append as is
                        checked_orders.append(item)
                    else:
                        raise Exception(f"Invalid order submitted: {item}")
            else:
                raise Exception(f"Invalid order/s submitted: '{orders}' received")

            return checked_orders

        def add_strategy_data(orders):
            # Append strategy parameters to each order
            for order in orders:
                order.instrument = (
                    self.instrument if not order.instrument else order.instrument
                )
                order.strategy = (
                    self._strategy.name
                    if "name" in self._strategy.__dict__
                    else self._strategy_name
                )
                order.granularity = self._strategy_params["granularity"]
                order._sizing = self._strategy_params["sizing"]
                order._risk_pc = self._strategy_params["risk_pc"]

        def check_order_details(orders: list) -> None:
            # Check details for order type have been provided
            for ix, order in enumerate(orders):
                order.instrument = (
                    order.instrument
                    if order.instrument is not None
                    else self.instrument
                )
                if order.order_type in ["market", "limit", "stop-limit", "reduce"]:
                    if not order.direction:
                        # Order direction was not provided, delete order
                        del orders[ix]
                        continue

                # Check that an exchange has been specified
                if order.exchange is None:
                    # Exchange not specified
                    if self._multiple_brokers:
                        # Trading across multiple venues
                        raise Exception(
                            "The exchange to which an order is to be "
                            + "submitted must be specified when trading across "
                            + "multiple venues. Please include the 'exchange' "
                            + "argument when creating an order."
                        )
                    else:
                        # Trading on single venue, auto fill
                        order.exchange = self._broker_name

        # Perform checks
        checked_orders = check_type(orders)
        add_strategy_data(checked_orders)
        check_order_details(checked_orders)

        return checked_orders

    def _qualify_orders(
        self, orders: list, current_bars: dict, quote_bars: dict
    ) -> None:
        """Passes price data to order to populate missing fields."""

        for order in orders:
            # Get relevant broker
            broker = self._brokers[order.exchange]

            # Fetch precision for instrument
            try:
                precision = broker._utils.get_precision(order.instrument)
            except Exception as e:
                # Print exception
                print("AutoTrader exception when qualifying order:", e)

                # Skip this order
                continue

            if self._feed != "none":
                # Get order price from current bars
                if self._req_liveprice:
                    # Fetch current price
                    liveprice_func = getattr(
                        self._get_data, f"_{self._feed.lower()}_liveprice"
                    )
                    last_price = liveprice_func(order)
                else:
                    # Fetch pseudo-current price
                    try:
                        # Use instrument
                        last_price = self._get_data._pseduo_liveprice(
                            last=current_bars[order.instrument].Close,
                            quote_price=quote_bars[order.instrument].Close,
                        )
                    except:
                        # Use data name
                        last_price = self._get_data._pseduo_liveprice(
                            last=current_bars[order.data_name].Close,
                            quote_price=quote_bars[order.data_name].Close,
                        )

                if order.order_type not in ["close", "reduce", "modify"]:
                    if order.direction < 0:
                        order_price = last_price["bid"]
                        HCF = last_price["negativeHCF"]
                    else:
                        order_price = last_price["ask"]
                        HCF = last_price["positiveHCF"]
                else:
                    # Close, reduce or modify order type, provide dummy inputs
                    order_price = last_price["ask"]
                    HCF = last_price["positiveHCF"]

            else:
                # Do not provide order price yet
                order_price = None
                HCF = None

            # Call order to update
            order(broker=broker, order_price=order_price, HCF=HCF, precision=precision)

    def _update_virtual_broker(self, current_bars: dict) -> None:
        """Updates virtual broker with latest price data."""
        # TODO - the conditional here should allow specifically updating by L1,
        # not only when feed=none
        if self._feed == "none":
            # None data feed provided, use L1 to update
            for instrument, brokers in self._instrument_to_broker.items():
                for broker in brokers:
                    broker._update_instrument(instrument)

        else:
            # Using OHLC data feed
            for product, bar in current_bars.items():
                brokers = self._instrument_to_broker[product]
                for broker in brokers:
                    broker._update_positions(instrument=product, candle=bar)

    def _create_trade_results(self, broker_histories: dict) -> dict:
        """Constructs bot-specific trade summary for post-processing."""
        trade_results = TradeAnalysis(self._broker, broker_histories, self.instrument)
        trade_results.indicators = (
            self._strategy.indicators if hasattr(self._strategy, "indicators") else None
        )
        trade_results.data = self.data
        trade_results.interval = self._strategy_params["granularity"]
        self.trade_results = trade_results

    def _get_iteration_range(self) -> int:
        """Checks mode of operation and returns data iteration range. For backtesting,
        the entire dataset is iterated over. For livetrading, only the latest candle
        is used. ONLY USED IN BACKTESTING NOW.
        """
        start_range = self._strategy_params["period"]
        end_range = len(self.data)

        if len(self.data) < start_range:
            raise Exception(
                "There are not enough bars in the data to "
                + "run the backtest with the current strategy "
                + "configuration settings. Either extend the "
                + "backtest period, or reduce the PERIOD key of "
                + "your strategy configuration."
            )

        return start_range, end_range

    @staticmethod
    def _check_ohlc_data(
        ohlc_data: pd.DataFrame,
        timestamp: datetime,
        indexing: str = "open",
        tail_bars: int = None,
        check_for_future_data: bool = True,
    ) -> pd.DataFrame:
        """Checks the index of inputted data to ensure it contains no future
        data.

        Parameters
        ----------
        ohlc_data : pd.DataFrame
            DESCRIPTION.
        timestamp : datetime
            The current timestamp.
        indexing : str, optional
            How the OHLC data has been indexed (either by bar 'open' time, or
            bar 'close' time). The default is 'open'.
        tail_bars : int, optional
            If provided, the data will be truncated to provide the number
            of bars specified. The default is None.
        check_for_future_data : bool, optional
            A flag to check for future entries in the data. The default is True.

        Raises
        ------
        Exception
            When an unrecognised data indexing type is specified.

        Returns
        -------
        past_data : pd.DataFrame
            The checked data.

        """
        if check_for_future_data:
            if indexing.lower() == "open":
                past_data = ohlc_data[ohlc_data.index < timestamp]
            elif indexing.lower() == "close":
                past_data = ohlc_data[ohlc_data.index <= timestamp]
            else:
                raise Exception(f"Unrecognised indexing type '{indexing}'.")
        else:
            past_data = ohlc_data

        if tail_bars is not None:
            past_data = past_data.tail(tail_bars)

        return past_data

    def _check_auxdata(
        self,
        auxdata: dict,
        timestamp: datetime,
        indexing: str = "open",
        tail_bars: int = None,
        check_for_future_data: bool = True,
    ) -> dict:
        """Function to check the strategy auxiliary data.

        Parameters
        ----------
        auxdata : dict
            The strategy's auxiliary data.
        timestamp : datetime
            The current timestamp.
        indexing : str, optional
            How the OHLC data has been indexed (either by bar 'open' time, or
            bar 'close' time). The default is 'open'.
        tail_bars : int, optional
            If provided, the data will be truncated to provide the number
            of bars specified. The default is None.
        check_for_future_data : bool, optional
            A flag to check for future entries in the data. The default is True.

        Returns
        -------
        dict
            The checked auxiliary data.
        """
        processed_auxdata = {}
        for key, item in auxdata.items():
            if isinstance(item, pd.DataFrame) or isinstance(item, pd.Series):
                processed_auxdata[key] = self._check_ohlc_data(
                    item, timestamp, indexing, tail_bars, check_for_future_data
                )
            else:
                processed_auxdata[key] = item
        return processed_auxdata

    def _check_data(self, timestamp: datetime, indexing: str = "open") -> dict:
        """Function to return trading data based on the current timestamp. If
        dynamc_data updates are required (eg. when livetrading), the
        datastream will be refreshed each update to retrieve new data. The
        data will then be checked to ensure that there is no future data
        included.

        Parameters
        ----------
        timestamp : datetime
            DESCRIPTION.
        indexing : str, optional
            DESCRIPTION. The default is 'open'.

        Returns
        -------
        strat_data : dict
            The checked strategy data.
        current_bars : dict(pd.core.series.Series)
            The current bars for each product.
        quote_bars : dict(pd.core.series.Series)
            The current quote data bars for each product.
        sufficient_data : bool
            Boolean flag whether sufficient data is available.

        """

        def get_current_bars(
            data: pd.DataFrame,
            quote_data: bool = False,
            processed_strategy_data: dict = None,
        ) -> dict:
            """Returns the current bars of data. If the inputted data is for
            quote bars, then the quote_data boolean will be True.
            """
            if len(data) > 0:
                current_bars = self.Stream.get_trading_bars(
                    data=data,
                    quote_bars=quote_data,
                    timestamp=timestamp,
                    processed_strategy_data=processed_strategy_data,
                )
            else:
                current_bars = None
            return current_bars

        def process_strat_data(original_strat_data, check_for_future_data):
            sufficient_data = True

            if isinstance(original_strat_data, dict):
                if "aux" in original_strat_data:
                    # Auxiliary data is being used
                    base_data = original_strat_data["base"]
                    processed_auxdata = self._check_auxdata(
                        original_strat_data["aux"],
                        timestamp,
                        indexing,
                        no_bars,
                        check_for_future_data,
                    )
                else:
                    # MTF data
                    base_data = original_strat_data

                # Process base OHLC data
                processed_basedata = {}
                if isinstance(base_data, dict):
                    # Base data is multi-timeframe; process each timeframe
                    for granularity, data in base_data.items():
                        processed_basedata[granularity] = self._check_ohlc_data(
                            data, timestamp, indexing, no_bars, check_for_future_data
                        )
                elif isinstance(base_data, pd.DataFrame) or isinstance(
                    base_data, pd.Series
                ):
                    # Base data is a timeseries already, check directly
                    processed_basedata = self._check_ohlc_data(
                        base_data, timestamp, indexing, no_bars, check_for_future_data
                    )

                # Combine the results of the conditionals above
                strat_data = {}
                if "aux" in original_strat_data:
                    strat_data["aux"] = processed_auxdata
                    strat_data["base"] = processed_basedata
                else:
                    strat_data = processed_basedata

                # Extract current bar
                first_tf_data = processed_basedata[list(processed_basedata.keys())[0]]
                current_bars = get_current_bars(
                    first_tf_data, processed_strategy_data=strat_data
                )

                # Check that enough bars have accumulated
                if len(first_tf_data) < no_bars:
                    sufficient_data = False

            elif isinstance(original_strat_data, pd.DataFrame):
                strat_data = self._check_ohlc_data(
                    original_strat_data,
                    timestamp,
                    indexing,
                    no_bars,
                    check_for_future_data,
                )
                current_bars = get_current_bars(
                    strat_data, processed_strategy_data=strat_data
                )

                # Check that enough bars have accumulated
                if len(strat_data) < no_bars:
                    sufficient_data = False

            elif original_strat_data is None:
                # Using none data
                strat_data = None
                current_bars = {}
                sufficient_data = True

            else:
                raise Exception("Unrecognised data type. Cannot process.")

            return strat_data, current_bars, sufficient_data

        # Define minimum number of bars for strategy to run
        no_bars = self._strategy_params["period"]

        if self._backtest_mode:
            check_for_future_data = True
            if self._dynamic_data:
                self._refresh_data(timestamp)
        else:
            # Livetrading
            self._refresh_data(timestamp)
            check_for_future_data = False

        strat_data, current_bars, sufficient_data = process_strat_data(
            self._strat_data, check_for_future_data
        )

        # Process quote data
        if isinstance(self.quote_data, dict):
            processed_quote_data = {}
            for instrument in self.quote_data:
                processed_quote_data[instrument] = self._check_ohlc_data(
                    self.quote_data[instrument],
                    timestamp,
                    indexing,
                    no_bars,
                    check_for_future_data,
                )
            quote_data = processed_quote_data[instrument]  # Dummy

        elif isinstance(self.quote_data, pd.DataFrame):
            quote_data = self._check_ohlc_data(
                self.quote_data, timestamp, indexing, no_bars, check_for_future_data
            )
            processed_quote_data = {self.instrument: quote_data}

        elif self.quote_data is None:
            # Using 'none' data feed
            quote_bars = current_bars
            return strat_data, current_bars, quote_bars, sufficient_data

        else:
            raise Exception("Unrecognised data type. Cannot process.")

        # Get quote bars
        quote_bars = get_current_bars(quote_data, True, processed_quote_data)

        return strat_data, current_bars, quote_bars, sufficient_data

    def _check_last_bar(self, current_bars: dict) -> bool:
        """Checks for new data to prevent duplicate signals."""
        if self._allow_duplicate_bars:
            new_data = True
        else:
            try:
                duplicated_bars = []
                for product, bar in current_bars.items():
                    if (bar == self._last_bars[product]).all():
                        duplicated_bars.append(True)
                    else:
                        duplicated_bars.append(False)

                if len(duplicated_bars) == sum(duplicated_bars):
                    new_data = False
                else:
                    new_data = True

            except:
                new_data = True

        # Reset last bars
        self._last_bars = current_bars

        if int(self._verbosity) > 1 and not new_data:
            print("Duplicate bar detected. Skipping.")

        return new_data

    def _check_strategy_for_plot_data(self, use_strat_plot_data: bool = False):
        """Checks the bot's strategy to see if it has the plot_data attribute.

        Returns
        -------
        plot_data : pd.DataFrame
            The data to plot.

        Notes
        -----
        This method is a placeholder for a future feature, allowing
        customisation of what is plotted by setting plot_data and plot_type
        attributes from within a strategy.
        """
        strat_params = self._strategy.__dict__
        if "plot_data" in strat_params and use_strat_plot_data:
            plot_data = strat_params["plot_data"]
        else:
            plot_data = self.data

        return plot_data

    def _strategy_shutdown(
        self,
    ):
        if self._strategy_shutdown_method is not None:
            try:
                shutdown_method = getattr(
                    self._strategy, self._strategy_shutdown_method
                )
                shutdown_method()
            except AttributeError:
                print(
                    f"\nShutdown method '{self._strategy_shutdown_method}' not found!"
                )

    def _replace_data(self, data: pd.DataFrame) -> None:
        """Function to replace the data assigned locally and to the strategy.
        Called when there is a mismatch in data lengths during multi-instrument
        backtests in periodic update mode.
        """
        self.data = data
        self._strategy.data = data

    @staticmethod
    def _submit_order(broker, order, *args, **kwargs):
        "The default order execution method."
        broker.place_order(order, *args, **kwargs)



from autotrader.brokers.trading import Order
from autotrader.brokers.broker import AbstractBroker
from autotrader.brokers.broker_utils import BrokerUtils

"""
Notes:
    - Public methods are called from outside the broker module, and so must
      retain functionality of input arguments. If necessary, they can simply
      be wrapper methods.
    - Private methods are broker-specific.
"""


class Broker(AbstractBroker):
    def __init__(self, config: dict, utils: BrokerUtils = None) -> None:
        """AutoTrader Broker Class constructor."""

        self._utils = utils if utils is not None else BrokerUtils()

        # Unpack config and connect to broker-side API

    def __repr__(self):
        return "AutoTrader Broker interface"

    def __str__(self):
        return "AutoTrader Broker interface"

    def get_NAV(self) -> float:
        """Returns the net asset/liquidation value of the account."""
        pass

    def get_balance(self) -> float:
        """Returns account balance."""
        pass

    def place_order(self, order: Order, **kwargs) -> None:
        """Disassemble order_details dictionary to place order."""
        # Call order to set order time
        order()

        # Submit order to broker

    def get_orders(self, instrument: str = None, **kwargs) -> dict:
        """Returns all pending orders (have not been filled) in the account."""
        pass

    def cancel_order(self, order_id: int, **kwargs) -> None:
        """Cancels order by order ID."""
        pass

    def get_trades(self, instrument: str = None, **kwargs) -> dict:
        """Returns the trades (fills) made by the account."""
        pass

    def get_positions(self, instrument: str = None, **kwargs) -> dict:
        """Gets the current positions open on the account.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument name (symbol). The default is None.

        Returns
        -------
        open_positions : dict
            A dictionary containing details of the open positions.
        """
        pass

    # Define here any private methods to support the public methods above



from autotrader.brokers.broker_utils import BrokerUtils


class Utils(BrokerUtils):
    def __init__(self, exchange: str = None, **kwargs):
        if exchange is not None:
            self.connect_to_exchange(exchange)

        # Stored instrument precisions
        self._instrument_precisions = {}

    def connect_to_exchange(self, exchange: str):
        try:
            import ccxt

            self.api = getattr(ccxt, exchange)()
            self.markets = self.api.load_markets()
        except ImportError:
            raise Exception("Please install ccxt to connect " + "to CCXT.")
        # Except network error?

    def __repr__(self):
        return "AutoTrader-CCXT Broker Utilities"

    def __str__(self):
        return "AutoTrader-CCXT Broker Utilities"

    def get_precision(self, instrument, *args, **kwargs):
        """Returns the precision of the instrument."""
        if instrument in self._instrument_precisions:
            # Precision already fetched, use stored value
            unified_response = self._instrument_precisions[instrument]
        else:
            # Fetch precision
            market = self._get_market(instrument)
            precision = market["precision"]

            size_precision = precision["amount"]
            price_precision = precision["price"]

            # Check for any decimals
            if "." in str(size_precision):
                size_precision = str(size_precision)[::-1].find(".")
            if "." in str(price_precision):
                price_precision = str(price_precision)[::-1].find(".")

            unified_response = {
                "size": size_precision,
                "price": price_precision,
            }

            # Store for later use
            self._instrument_precisions[instrument] = unified_response

        return unified_response

    def _get_market(self, instrument, *args, **kwargs):
        """Returns the raw get_market response from a CCXT exchange"""
        if instrument in self.markets:
            market = self.markets[instrument]
        elif instrument.split(":")[0] in self.markets:
            market = self.markets[instrument.split(":")[0]]
        elif f"{instrument.split('USDT')[0]}/USDT" in self.markets:
            market = self.markets[f"{instrument.split('USDT')[0]}/USDT"]
        else:
            raise Exception(
                f"{instrument} does not appear to be listed. "
                + "Please double check the naming."
            )
        return market

    def get_stepsize(self, instrument, *args, **kwargs):
        """Returns the stepsize for an instrument."""
        market = self._get_market(instrument)
        stepsize = float(market["limits"]["amount"]["min"])
        return stepsize

    def get_min_notional(self, instrument, *args, **kwargs):
        """Returns the minimum notional value a trade should hold."""
        market = self._get_market(instrument)
        min_notional = float(market["limits"]["cost"]["min"])
        return min_notional

    def get_ticksize(self, instrument, *args, **kwargs):
        """Returns the ticksize for an instrument."""
        market = self._get_market(instrument)
        try:
            ticksize = float(market["info"]["filters"][0]["tickSize"])
        except:
            raise Exception("Cannot retrieve ticksize.")
        return ticksize



from abc import ABC, abstractmethod
from autotrader.brokers.trading import Order
from autotrader.brokers.broker_utils import BrokerUtils


class AbstractBroker(ABC):
    @abstractmethod
    def __init__(self, config: dict, utils: BrokerUtils = None) -> None:
        """AutoTrader Broker Class constructor."""
        pass

    @abstractmethod
    def __repr__(self):
        return "AutoTrader Broker interface"

    @abstractmethod
    def __str__(self):
        return "AutoTrader Broker interface"

    @abstractmethod
    def get_NAV(self, *args, **kwargs) -> float:
        """Returns the net asset/liquidation value of the account."""
        pass

    @abstractmethod
    def get_balance(self, *args, **kwargs) -> float:
        """Returns account balance."""
        pass

    @abstractmethod
    def place_order(self, order: Order, *args, **kwargs) -> None:
        """Translate order and place via exchange API."""
        pass

    @abstractmethod
    def get_orders(self, instrument: str = None, *args, **kwargs) -> dict:
        """Returns all pending orders (have not been filled) in the account."""
        pass

    @abstractmethod
    def cancel_order(self, order_id: int, *args, **kwargs) -> None:
        """Cancels order by order ID."""
        pass

    @abstractmethod
    def get_trades(self, instrument: str = None, *args, **kwargs) -> dict:
        """Returns the trades (fills) made by the account."""
        pass

    @abstractmethod
    def get_positions(self, instrument: str = None, *args, **kwargs) -> dict:
        """Gets the current positions open on the account.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument name (symbol). The default is None.

        Returns
        -------
        open_positions : dict
            A dictionary containing details of the open positions.
        """
        pass



from autotrader.brokers.broker_utils import BrokerUtils


class Utils(BrokerUtils):
    def __init__(self, **kwargs):
        pass

    def __repr__(self):
        return "AutoTrader Broker Utilities"

    def __str__(self):
        return "AutoTrader Broker Utilities"



from distutils.command.config import config
import os
import time
import pandas as pd
from typing import Union
from decimal import Decimal
from autotrader.brokers.trading import Order
from datetime import datetime, timedelta, timezone
from autotrader.brokers.broker_utils import OrderBook
import sys


class AutoData:
    """AutoData class to retrieve price data.

    Attributes
    ----------
    _home_curreny : str
        the home currency of the account (used for retrieving quote data)

    _allow_dancing_bears : bool
        Allow incomplete candlesticks in data retrieval.

    """

    def __init__(
        self,
        data_config: dict = None,
        allow_dancing_bears: bool = False,
        home_currency: str = None,
        **kwargs,
    ) -> None:
        """Instantiates AutoData.

        Parameters
        ----------
        data_config : dict, optional
            The configuration dictionary for the data source to be used. This
            is created automatically in autotrader.utilities.get_data_config.
            The default is None.
        allow_dancing_bears : bool, optional
            A flag to allow incomplete bars to be returned in the data. The
            default is False.
        home_currency : str, optional
            The home currency to use when fetching quote data. The default
            is None.
        kwargs : optional

        Returns
        -------
        None
            AutoData will be instantiated and ready to fetch price data.
        """
        # Merge kwargs and data_config
        if data_config is None and kwargs is not None:
            data_config = {}
        for key, item in kwargs.items():
            data_config.setdefault(key, item)

        def configure_local_feed(data_config):
            """Configures the attributes for a local data feed."""
            self._feed = "local"
            self.api = None
            self._data_directory = (
                data_config["data_dir"] if "data_dir" in data_config else None
            )
            self._spread_units = (
                data_config["spread_units"]
                if "spread_units" in data_config
                else "percentage"
            )
            self._spread = data_config["spread"] if "spread" in data_config else 0

        if not data_config:
            configure_local_feed({})
        else:
            self._feed = data_config["data_source"].lower()

            if data_config["data_source"].lower() == "oanda":
                API = data_config["API"]
                ACCESS_TOKEN = data_config["ACCESS_TOKEN"]
                self.ACCOUNT_ID = data_config["ACCOUNT_ID"]
                port = data_config["PORT"]

                try:
                    import v20

                    self.api = v20.Context(hostname=API, token=ACCESS_TOKEN, port=port)
                except ImportError:
                    raise Exception("Please install v20 to use the Oanda data feed.")

            elif data_config["data_source"].lower() == "ib":
                host = data_config["host"]
                port = data_config["port"]
                client_id = data_config["clientID"] + 1
                read_only = data_config["read_only"]
                account = data_config["account"]

                try:
                    import ib_insync
                    from autotrader.brokers.ib.utils import Utils as IB_Utils

                    self.api = ib_insync.IB()
                    self.IB_Utils = IB_Utils
                    self.api.connect(
                        host=host,
                        port=port,
                        clientId=client_id,
                        readonly=read_only,
                        account=account,
                    )
                except ImportError:
                    raise Exception("Please install ib_insync to use the IB data feed.")

            elif data_config["data_source"].lower() == "ccxt":
                try:
                    import ccxt

                    self._ccxt_exchange = data_config["exchange"]

                    if "api" in kwargs:
                        # Use API instance provided
                        self.api = kwargs["api"]

                    else:
                        # Check if exchange options were provided
                        if "config" in data_config:
                            # Use config dictionary provided directly
                            ccxt_config = data_config["config"]
                        elif "secret" in data_config and "api_key" in data_config:
                            # Create config dict with api key and secret
                            ccxt_config = {
                                "secret": data_config["secret"],
                                "apiKey": data_config["api_key"],
                            }
                        else:
                            # Create empty config dict
                            ccxt_config = {}

                        # Add any other keys to the config
                        extra_keys = ["options", "password"]
                        for key in extra_keys:
                            if key in data_config:
                                ccxt_config[key] = data_config[key]

                        # Create CCXT instance
                        exchange_module = getattr(ccxt, data_config["exchange"])
                        self.api = exchange_module(config=ccxt_config)

                        # Check sandbox mode
                        if "sandbox_mode" in data_config:
                            self.api.set_sandbox_mode(data_config["sandbox_mode"])

                        # Load markets
                        markets = self.api.load_markets()

                except ImportError:
                    raise Exception("Please install ccxt to use the CCXT data feed.")

            elif data_config["data_source"].lower() == "dydx":
                try:
                    from dydx3 import Client

                    self.api = Client(host="https://api.dydx.exchange")
                except ImportError:
                    raise Exception(
                        "Please install dydx-v3-python to use the dydx data feed."
                    )

            elif data_config["data_source"].lower() == "yahoo":
                try:
                    import yfinance as yf

                    self.api = yf.download
                except ImportError:
                    raise Exception(
                        "Please install yfinance to use "
                        + "the Yahoo Finance data feed."
                    )

            elif data_config["data_source"].lower() == "local":
                configure_local_feed(data_config)

            elif data_config["data_source"].lower() == "none":
                # No data feed required
                self.api = None

            else:
                raise Exception(f"Unknown data source '{self._feed}'.")

        self._allow_dancing_bears = allow_dancing_bears
        self._home_currency = home_currency

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        feed_str = self._ccxt_exchange if self._feed == "ccxt" else self._feed
        return f"AutoData ({feed_str} feed)"

    def fetch(
        self,
        instrument: str,
        granularity: str = None,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Unified data retrieval api."""
        func = getattr(self, f"_{self._feed}")
        data = func(
            instrument,
            granularity=granularity,
            count=count,
            start_time=start_time,
            end_time=end_time,
            *args,
            **kwargs,
        )
        return data

    def _quote(self, *args, **kwargs):
        """Unified quote data retrieval api."""
        func = getattr(self, f"_{self._feed}_quote_data")
        data = func(*args, **kwargs)
        return data

    def L1(self, instrument=None, *args, **kwargs):
        """Unified level 1 data retrieval api."""
        # Get orderbook
        orderbook = self.L2(instrument, *args, **kwargs)

        # Construct response
        response = {
            "bid": orderbook.bids["price"][0],
            "ask": orderbook.asks["price"][0],
            "bid_size": orderbook.bids["size"][0],
            "ask_size": orderbook.asks["size"][0],
        }

        return response

    def L2(self, instrument=None, *args, **kwargs):
        """Unified level 2 data retrieval api."""
        func = getattr(self, f"_{self._feed}_orderbook")
        data = func(instrument, *args, **kwargs)
        if self._feed == "local":
            book = data
        else:
            book = OrderBook(instrument, data)
        return book

    def trades(self, instrument, *args, **kwargs):
        func = getattr(self, f"_{self._feed}_trades")
        trades = func(instrument, *args, **kwargs)
        return trades

    def _oanda(
        self,
        instrument: str,
        granularity: str,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
    ) -> pd.DataFrame:
        """Retrieves historical price data of a instrument from Oanda v20 API.

        Parameters
        ----------
        instrument : str
            The instrument to fetch data for.
        granularity : str
            The candlestick granularity, specified as a TimeDelta string
            (eg. '30s', '5min' or '1d').
        count : int, optional
            The number of candles to fetch (maximum 5000). The default is None.
        start_time : datetime, optional
            The data start time. The default is None.
        end_time : datetime, optional
            The data end time. The default is None.

        Returns
        -------
        data : DataFrame
            The price data, as an OHLC DataFrame.

        Notes
        -----
            If a candlestick count is provided, only one of start time or end
            time should be provided. If neither is provided, the N most
            recent candles will be provided. If both are provided, the count
            will be ignored, and instead the dates will be used.
        """
        gran_map = {
            5: "S5",
            10: "S10",
            15: "S15",
            30: "S30",
            60: "M1",
            120: "M2",
            240: "M4",
            300: "M5",
            600: "M10",
            900: "M15",
            1800: "M30",
            3600: "H1",
            7200: "H2",
            10800: "H3",
            14400: "H4",
            21600: "H6",
            28800: "H8",
            43200: "H12",
            86400: "D",
            604800: "W",
            2419200: "M",
        }
        granularity = gran_map[pd.Timedelta(granularity).total_seconds()]

        if count is not None:
            # either of count, start_time+count, end_time+count (or start_time+end_time+count)
            # if count is provided, count must be less than 5000
            if start_time is None and end_time is None:
                # fetch count=N most recent candles
                response = self.api.instrument.candles(
                    instrument, granularity=granularity, count=count
                )
                data = self._response_to_df(response)

            elif start_time is not None and end_time is None:
                # start_time + count
                from_time = start_time.timestamp()
                response = self.api.instrument.candles(
                    instrument, granularity=granularity, count=count, fromTime=from_time
                )
                data = self._response_to_df(response)

            elif end_time is not None and start_time is None:
                # end_time + count
                to_time = end_time.timestamp()
                response = self.api.instrument.candles(
                    instrument, granularity=granularity, count=count, toTime=to_time
                )
                data = self._response_to_df(response)

            else:
                # start_time+end_time+count
                # print("Warning: ignoring count input since start_time and",
                #        "end_time times have been specified.")
                from_time = start_time.timestamp()
                to_time = end_time.timestamp()

                # try to get data
                response = self.api.instrument.candles(
                    instrument,
                    granularity=granularity,
                    fromTime=from_time,
                    toTime=to_time,
                )

                # If the request is rejected, max candles likely exceeded
                if response.status != 200:
                    data = self._get_extended_oanda_data(
                        instrument, granularity, from_time, to_time
                    )
                else:
                    data = self._response_to_df(response)

        else:
            # count is None
            # Assume that both start_time and end_time have been specified.
            from_time = start_time.timestamp()
            to_time = end_time.timestamp()

            # try to get data
            response = self.api.instrument.candles(
                instrument, granularity=granularity, fromTime=from_time, toTime=to_time
            )

            # If the request is rejected, max candles likely exceeded
            if response.status != 200:
                data = self._get_extended_oanda_data(
                    instrument, granularity, from_time, to_time
                )
            else:
                data = self._response_to_df(response)

        return data

    def _oanda_liveprice(self, order: Order, **kwargs) -> dict:
        """Returns current price (bid+ask) and home conversion factors."""
        response = self.api.pricing.get(
            accountID=self.ACCOUNT_ID, instruments=order.instrument
        )
        ask = response.body["prices"][0].closeoutAsk
        bid = response.body["prices"][0].closeoutBid
        negativeHCF = response.body["prices"][
            0
        ].quoteHomeConversionFactors.negativeUnits
        positiveHCF = response.body["prices"][
            0
        ].quoteHomeConversionFactors.positiveUnits

        price = {
            "ask": ask,
            "bid": bid,
            "negativeHCF": negativeHCF,
            "positiveHCF": positiveHCF,
        }

        return price

    def _oanda_orderbook(self, instrument, time=None, *args, **kwargs):
        """Returns the orderbook from Oanda."""
        response = self.api.pricing.get(
            accountID=self.ACCOUNT_ID, instruments=instrument
        )
        prices = response.body["prices"][0].dict()

        # Unify format
        orderbook = {}
        for side in ["bids", "asks"]:
            orderbook[side] = []
            for level in prices[side]:
                orderbook[side].append(
                    {"price": level["price"], "size": level["liquidity"]}
                )
        return orderbook

    def _get_extended_oanda_data(self, instrument, granularity, from_time, to_time):
        """Returns historical data between a date range."""
        max_candles = 5000

        my_int = self._granularity_to_seconds(granularity, "oanda")
        end_time = to_time - my_int
        partial_from = from_time
        response = self.api.instrument.candles(
            instrument,
            granularity=granularity,
            fromTime=partial_from,
            count=max_candles,
        )
        data = self._response_to_df(response)
        last_time = data.index[-1].timestamp()

        while last_time < end_time:
            candles = min(max_candles, int((end_time - last_time) / my_int))
            partial_from = last_time
            response = self.api.instrument.candles(
                instrument,
                granularity=granularity,
                fromTime=partial_from,
                count=candles,
            )

            partial_data = self._response_to_df(response)
            data = pd.concat([data, partial_data])
            last_time = data.index[-1].timestamp()

        return data

    def _oanda_quote_data(
        self,
        data: pd.DataFrame,
        pair: str,
        granularity: str,
        start_time: datetime,
        end_time: datetime,
        count: int = None,
    ):
        """Function to retrieve price conversion data."""
        gran_map = {
            5: "S5",
            10: "S10",
            15: "S15",
            30: "S30",
            60: "M1",
            120: "M2",
            240: "M4",
            300: "M5",
            600: "M10",
            900: "M15",
            1800: "M30",
            3600: "H1",
            7200: "H2",
            10800: "H3",
            14400: "H4",
            21600: "H6",
            28800: "H8",
            43200: "H12",
            86400: "D",
            604800: "W",
            2419200: "M",
        }
        granularity = gran_map[pd.Timedelta(granularity).total_seconds()]

        base_currency = pair[:3]
        quote_currency = pair[-3:]

        if self._home_currency is None or quote_currency == self._home_currency:
            # Use data as quote data
            quote_data = data

        else:
            if self._home_currency == base_currency:
                # Disturb the data by machine precision to prompt HCF calculation
                quote_data = (1 + 1e-15) * data

            else:
                # Try download quote data
                conversion_pair = self._home_currency + "_" + quote_currency
                if conversion_pair == pair:
                    # Same instrument
                    quote_data = data

                else:
                    # Different instrument, fetch data
                    try:
                        # Directly
                        quote_data = self.oanda(
                            instrument=conversion_pair,
                            granularity=granularity,
                            start_time=start_time,
                            end_time=end_time,
                            count=count,
                        )
                    except:
                        # Failed
                        try:
                            # Invert conversion pair
                            conversion_pair = quote_currency + "_" + self._home_currency
                            inverse_quote_data = self.oanda(
                                instrument=conversion_pair,
                                granularity=granularity,
                                start_time=start_time,
                                end_time=end_time,
                                count=count,
                            )
                            quote_data = (
                                1 / inverse_quote_data[["Open", "High", "Low", "Close"]]
                            )

                        except:
                            # Failed, just used original data
                            quote_data = data

        return quote_data

    def _check_oanda_response(self, response):
        """Placeholder method to check Oanda API response."""
        if response.status != 200:
            print(response.reason)

    def _response_to_df(self, response):
        """Function to convert api response into a pandas dataframe."""
        try:
            candles = response.body["candles"]
        except KeyError:
            raise Exception(
                "Error dowloading data - please check instrument"
                + " format and try again."
            )

        times = []
        close_price, high_price, low_price, open_price, volume = [], [], [], [], []

        if self._allow_dancing_bears:
            # Allow all candles
            for candle in candles:
                times.append(candle.time)
                close_price.append(float(candle.mid.c))
                high_price.append(float(candle.mid.h))
                low_price.append(float(candle.mid.l))
                open_price.append(float(candle.mid.o))
                volume.append(float(candle.volume))

        else:
            # Only allow complete candles
            for candle in candles:
                if candle.complete:
                    times.append(candle.time)
                    close_price.append(float(candle.mid.c))
                    high_price.append(float(candle.mid.h))
                    low_price.append(float(candle.mid.l))
                    open_price.append(float(candle.mid.o))
                    volume.append(float(candle.volume))

        dataframe = pd.DataFrame(
            {
                "Open": open_price,
                "High": high_price,
                "Low": low_price,
                "Close": close_price,
                "Volume": volume,
            }
        )
        dataframe.index = pd.to_datetime(times)
        dataframe.drop_duplicates(inplace=True)

        return dataframe

    @staticmethod
    def _granularity_to_seconds(granularity: str, feed: str):
        """Converts the granularity to time in seconds."""
        if feed.lower() == "oanda":
            allowed_granularities = (
                "S5",
                "S10",
                "S15",
                "S30",
                "M1",
                "M2",
                "M4",
                "M5",
                "M10",
                "M15",
                "M30",
                "H1",
                "H2",
                "H3",
                "H4",
                "H6",
                "H8",
                "H12",
                "D",
                "W",
                "M",
            )

            if granularity not in allowed_granularities:
                raise Exception(
                    f"Invalid granularity '{granularity}' for " + "{feed} data feed."
                )

            letter = granularity[0]

            if len(granularity) > 1:
                number = float(granularity[1:])
            else:
                number = 1

            conversions = {"S": 1, "M": 60, "H": 60 * 60, "D": 60 * 60 * 24}

            my_int = conversions[letter] * number

        elif feed.lower() == "yahoo":
            # Note: will not work for week or month granularities

            letter = granularity[-1]
            number = float(granularity[:-1])

            conversions = {"m": 60, "h": 60 * 60, "d": 60 * 60 * 24}

            my_int = conversions[letter] * number

        return my_int

    def _yahoo(
        self,
        instrument: str,
        granularity: str = None,
        count: int = None,
        start_time: str = None,
        end_time: str = None,
    ) -> pd.DataFrame:
        """Retrieves historical price data from yahoo finance.

        Parameters
        ----------
        instrument : str
            Ticker to dowload data for.
        granularity : str, optional
            The candlestick granularity. The default is None.
        count : int, optional
            The number of bars to fetch. The default is None.
        start_time : str, optional
            The start time as YYYY-MM-DD string or datetime object. The default
            is None.
        end_time : str, optional
            The end_time as YYYY-MM-DD string or datetime object. The default
            is None.

        Returns
        -------
        data : pd.DataFrame
            The price data, as an OHLC DataFrame.

        Notes
        -----
        If you are encountering a JSON error when using the yahoo finance API,
        try updating by running: pip install yfinance --upgrade --no-cache-dir

        Intraday data cannot exceed 60 days.
        """
        gran_map = {
            60: "1m",
            120: "2m",
            300: "5m",
            900: "15m",
            1800: "30m",
            3600: "60m",
            5400: "90m",
            3600: "1h",
            86400: "1d",
            432000: "5d",
            604800: "1wk",
            2419200: "1mo",
            7257600: "3mo",
        }
        granularity = gran_map[pd.Timedelta(granularity).total_seconds()]

        if count is not None and start_time is None and end_time is None:
            # Convert count to start and end dates (assumes end=now)
            end_time = datetime.now()
            start_time = end_time - timedelta(
                seconds=self._granularity_to_seconds(granularity, "yahoo") * count
            )

        data = self.api(
            tickers=instrument, start=start_time, end=end_time, interval=granularity
        )

        if data.index.tzinfo is None:
            # Data is naive, add UTC timezone
            data.index = data.index.tz_localize(timezone.utc)

        return data

    def _yahoo_orderbook(self, *args, **kwargs):
        raise Exception("Orderbook data is not available from Yahoo Finance.")

    def _yahoo_quote_data(
        self,
        data: pd.DataFrame,
        pair: str,
        interval: str,
        from_date: datetime,
        to_date: datetime,
        count: int = None,
    ):
        """Returns nominal price data - quote conversion not supported for
        Yahoo finance API.
        """
        return data

    def _check_IB_connection(self):
        """Checks if there is an active connection to IB."""
        self.api.sleep(0)
        connected = self.api.isConnected()
        if not connected:
            raise ConnectionError("No active connection to IB.")

    def _ib(
        self,
        instrument: str,
        granularity: str,
        count: int,
        start_time: datetime = None,
        end_time: datetime = None,
        order: Order = None,
        durationStr: str = "10 mins",
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Fetches data from IB.

        Parameters
        ----------
        instrument : str
            The product being traded.
        granularity : str
            The granularity of the price bars.
        count : int
            The number of bars to fetch.
        start_time : datetime, optional
            The data start time. The default is None.
        end_time : datetime, optional
            The data end time. The default is None.
        order : Order, optional
            The order object. The default is None.

        Raises
        ------
        NotImplementedError
            DESCRIPTION.

        Returns
        -------
        df : TYPE
            DESCRIPTION.

        Warnings
        --------
        This method is not recommended due to its high API poll rate.

        References
        ----------
        https://ib-insync.readthedocs.io/api.html?highlight=reqhistoricaldata#
        """
        raise NotImplementedError(
            "Historical market data from IB is not yet supported."
        )
        # TODO - implement
        contract = IB_Utils.build_contract(order)

        dt = ""
        barsList = []
        while True:
            bars = self.api.reqHistoricalData(
                contract,
                endDateTime=dt,
                durationStr=durationStr,
                barSizeSetting=granularity,
                whatToShow="MIDPOINT",
                useRTH=True,
                formatDate=1,
            )
            if not bars:
                break
            barsList.append(bars)
            dt = bars[0].date

        # Convert bars to DataFrame
        allBars = [b for bars in reversed(barsList) for b in bars]
        df = self.api.util.df(allBars)
        return df

    def _ib_liveprice(self, order: Order, snapshot: bool = False, **kwargs) -> dict:
        """Returns current price (bid+ask) and home conversion factors.

        Parameters
        ----------
        order: Order
            The AutoTrader Order.
        snapshot : bool, optional
            Request a snapshot of the price. The default is False.

        Returns
        -------
        dict
            A dictionary containing the bid and ask prices.
        """
        self._check_IB_connection()
        contract = self.IB_Utils.build_contract(order)
        self.api.qualifyContracts(contract)
        ticker = self.api.reqMktData(contract, snapshot=snapshot)
        while ticker.last != ticker.last:
            self.api.sleep(1)
        self.api.cancelMktData(contract)
        price = {
            "ask": ticker.ask,
            "bid": ticker.bid,
            "negativeHCF": 1,
            "positiveHCF": 1,
        }
        return price

    @staticmethod
    def _pseduo_liveprice(last: float, quote_price: float = None) -> dict:
        """Returns an artificial live bid and ask price, plus conversion
        factors.

        Parameters
        ----------
        last : float
            The last price of the product being traded.
        quote_price : float, optional
            The quote price of the product being traded against the account
            home currency. The default is None.

        Returns
        -------
        dict
            DESCRIPTION.

        """
        # TODO - build bid/ask spread into here, review virtual broker
        if quote_price is not None:
            # Use quote price to determine HCF
            if last == quote_price:
                # Quote currency matches account home currency
                negativeHCF = 1
                positiveHCF = 1
            else:
                # Quote data
                negativeHCF = 1 / quote_price
                positiveHCF = 1 / quote_price

        else:
            # No quote price provided
            negativeHCF = 1
            positiveHCF = 1

        price = {
            "ask": last,
            "bid": last,
            "negativeHCF": negativeHCF,
            "positiveHCF": positiveHCF,
        }

        return price

    def _local(
        self,
        instrument: str,
        start_time: Union[str, datetime] = None,
        end_time: Union[str, datetime] = None,
        utc: bool = True,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Reads and returns local price data.

        Parameters
        ----------
        instrument : str
            Either the absolute filepath of the local price data file, or, if
            a data directory was provided in the data_config dictionary, simply
            the name of the data file.
        start_time : str | datetime, optional
            The data start date. The default is None.
        end_time : str | datetime, optional
            The data end date. The default is None.
        utc : bool, optional
            Localise data to UTC. The default is True.

        Returns
        -------
        data : pd.DataFrame
            The price data, as an OHLC DataFrame.
        """
        filepath = (
            instrument
            if self._data_directory is None
            else os.path.join(self._data_directory, instrument)
        )

        data = pd.read_csv(filepath, index_col=0)
        data.index = pd.to_datetime(data.index, utc=utc)

        if start_time is not None and end_time is not None:
            # Filter by date range
            data = AutoData._check_data_period(data, start_time, end_time)

        return data

    def _local_orderbook(self, instrument=None, *args, **kwargs):
        """Returns an artificial orderbook based on the last bar of
        local price data.

        Parameters
        ----------
        instrument : str, optional
            The filename (absolute, or relative if data_dir was provided with
            data_config dictionary) of the instrument data. Only required if
            'midprice' argument is not provided.
        spread : float, optional
            The bid/ask spread value.
        spread_units : float, optional
            The units to which the spread refers to. Can be either 'price' or
            'percentage'.
        midprice : float, optional
            The midprice to use as a reference price.
        """
        spread_units = (
            kwargs["spread_units"] if "spread_units" in kwargs else self._spread_units
        )
        spread = kwargs["spread"] if "spread" in kwargs else self._spread

        # Get latest price
        if "midprice" in kwargs:
            midprice = kwargs["midprice"]
        else:
            # Load from OHLC data
            data = self._local(instrument)
            midprice = data.iloc[-1].Close

        if spread_units == "price":
            bid = midprice - 0.5 * spread
            ask = midprice + 0.5 * spread
        elif spread_units == "percentage":
            bid = midprice * (1 - 0.5 * spread / 100)
            ask = midprice * (1 + 0.5 * spread / 100)

        # Quantize
        bid = Decimal(bid).quantize(Decimal(str(midprice)))
        ask = Decimal(ask).quantize(Decimal(str(midprice)))

        # TODO - ability to add levels by book parameters
        data = {
            "bids": [
                {"price": bid, "size": 1e100},
            ],
            "asks": [
                {"price": ask, "size": 1e100},
            ],
        }
        orderbook = OrderBook(instrument, data)

        return orderbook

    def _local_quote_data(
        self,
        data: pd.DataFrame,
        pair: str,
        granularity: str,
        start_time: datetime,
        end_time: datetime,
        count: int = None,
    ):
        """Returns the original price data for a local data feed."""
        return data

    @staticmethod
    def _check_data_period(
        data: pd.DataFrame, start_date: datetime, end_date: datetime
    ) -> pd.DataFrame:
        """Checks and returns the dataset matching the backtest start and
        end dates (as close as possible).
        """
        return data[(data.index >= start_date) & (data.index <= end_date)]

    def _ccxt(
        self,
        instrument: str,
        granularity: str,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Retrieves historical price data of a instrument from an exchange
        instance of the CCXT package.

        Parameters
        ----------
        instrument : str
            The instrument to fetch data for.
        granularity : str
            The candlestick granularity (eg. "1m", "15m", "1h", "1d").
        count : int, optional
            The number of candles to fetch (maximum 5000). The default is None.
        start_time : datetime, optional
            The data start time. The default is None.
        end_time : datetime, optional
            The data end time. The default is None.

        Returns
        -------
        data : DataFrame
            The price data, as an OHLCV DataFrame.

        """

        # Check requested start and end times
        if end_time is not None and end_time > datetime.now(tz=end_time.tzinfo):
            raise Exception("End time cannot be later than the current time.")

        if start_time is not None and start_time > datetime.now(tz=start_time.tzinfo):
            raise Exception("Start time cannot be later than the current time.")

        if start_time is not None and end_time is not None and start_time > end_time:
            raise Exception("Start time cannot be later than the end time.")

        # Check granularity was provided
        if granularity is None:
            raise Exception("Please specify candlestick granularity.")

        def fetch_between_dates():
            # Fetches data between two dates
            max_count = 1000
            start_ts = int(start_time.timestamp() * 1000)
            end_ts = int(end_time.timestamp() * 1000)

            data = []
            while start_ts < end_ts:
                count = min(
                    max_count,
                    1
                    + (end_ts - start_ts)
                    / pd.Timedelta(granularity).total_seconds()
                    / 1000,
                )
                raw_data = self.api.fetchOHLCV(
                    instrument,
                    timeframe=granularity,
                    since=start_ts,
                    limit=int(count),
                    params=kwargs,
                )
                # Append data
                data += raw_data

                # Increment start_ts
                start_ts = raw_data[-1][0]

                # Sleep for API limit
                time.sleep(1)

            return data

        if count is not None:
            if start_time is None and end_time is None:
                # Fetch N most recent candles
                raw_data = self.api.fetchOHLCV(
                    instrument, timeframe=granularity, limit=count, params=kwargs
                )
            elif start_time is not None and end_time is None:
                # Fetch N candles since start_time
                start_ts = (
                    None if start_time is None else int(start_time.timestamp() * 1000)
                )
                raw_data = self.api.fetchOHLCV(
                    instrument,
                    timeframe=granularity,
                    since=start_ts,
                    limit=count,
                    params=kwargs,
                )
            elif end_time is not None and start_time is None:
                raise Exception(
                    "Fetching data from end_time and count is " + "not yet supported."
                )
            else:
                raw_data = fetch_between_dates()

        else:
            # Count is None
            try:
                assert start_time is not None and end_time is not None
                raw_data = fetch_between_dates()

            except AssertionError:
                raise Exception(
                    "When no count is provided, both start_time "
                    + "and end_time must be provided."
                )

        # Process data
        data = pd.DataFrame(
            raw_data, columns=["time", "Open", "High", "Low", "Close", "Volume"]
        ).set_index("time")
        data.index = pd.to_datetime(data.index, unit="ms")

        return data

    def _ccxt_quote_data(
        self,
        data: pd.DataFrame,
        pair: str = None,
        granularity: str = None,
        start_time: datetime = None,
        end_time: datetime = None,
        count: int = None,
    ):
        """Returns the original price data for a CCXT data feed."""
        return data

    def _ccxt_orderbook(self, instrument, limit=None, *args, **kwargs):
        """Returns the orderbook from a CCXT supported exchange."""
        response = self.api.fetchOrderBook(symbol=instrument)

        # Unify format
        orderbook = {}
        for side in ["bids", "asks"]:
            orderbook[side] = []
            for level in response[side]:
                orderbook[side].append({"price": level[0], "size": level[1]})
        return orderbook

    def _ccxt_trades(self, instrument):
        """Returns public trades from a CCXT exchange."""
        ccxt_trades = self.api.fetchTrades(instrument)

        # Convert to standard form
        trades = []
        for trade in ccxt_trades:
            unified_trade = {
                "direction": 1 if trade["side"] == "buy" else -1,
                "price": float(trade["price"]),
                "size": float(trade["amount"]),
                "time": datetime.fromtimestamp(trade["timestamp"] / 1000),
            }
            trades.append(unified_trade)

        return trades

    def _ccxt_funding_rate(self, instrument: str):
        """Returns the current funding rate."""
        response = self.api.fetchFundingRate(instrument)

        fr_dict = {
            "symbol": instrument,
            "rate": response["fundingRate"],
            "time": response["fundingDatetime"],
        }
        return fr_dict

    def _ccxt_funding_history(
        self,
        instrument: str,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
        params: dict = {},
    ):
        """Fetches the funding rate history."""

        def response2df(response):
            """Converts response to DataFrame."""
            times = []
            rates = []
            for chunk in response:
                times.append(pd.Timestamp(chunk["timestamp"], unit="ms"))
                rates.append(chunk["fundingRate"])
            return pd.DataFrame(data={"rate": rates}, index=times)

        def fetch_between_dates():
            # Fetches data between two dates
            count = 500
            start_ts = int(start_time.timestamp() * 1000)
            end_ts = int(end_time.timestamp() * 1000)

            rate_hist = pd.DataFrame()
            while start_ts <= end_ts:
                response = self.api.fetchFundingRateHistory(
                    symbol=instrument, since=start_ts, limit=count, params=params
                )

                # Append results
                df = response2df(response)
                rate_hist = pd.concat([rate_hist, df])

                # Increment start_ts
                start_ts = int(df.index[-1].timestamp() * 1000)

                # Sleep for API limit
                time.sleep(1)

            return rate_hist

        if count is not None:
            if start_time is None and end_time is None:
                # Fetch N most recent candles
                response = self.api.fetchFundingRateHistory(
                    symbol=instrument, limit=count, params=params
                )
                rate_hist = response2df(response)
            elif start_time is not None and end_time is None:
                # Fetch N candles since start_time
                start_ts = (
                    None if start_time is None else int(start_time.timestamp() * 1000)
                )
                response = self.api.fetchFundingRateHistory(
                    symbol=instrument, since=start_ts, limit=count, params=params
                )
                rate_hist = response2df(response)
            elif end_time is not None and start_time is None:
                raise Exception(
                    "Fetching data from end_time and count is " + "not yet supported."
                )
            else:
                rate_hist = fetch_between_dates()

        else:
            # Count is None
            if start_time is not None and end_time is not None:
                rate_hist = fetch_between_dates()
            else:
                response = self.api.fetchFundingRateHistory(
                    symbol=instrument,
                    params=params,
                )
                rate_hist = response2df(response)

        return rate_hist

    def _dydx(
        self,
        instrument: str,
        granularity: str,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Retrieves historical price data of a instrument from dYdX.

        Parameters
        ----------
        instrument : str
            The instrument/market to fetch data for.
        granularity : str
            The candlestick granularity (1DAY, 4HOURS, 1HOUR, 30MINS, 15MINS,
                                         5MINS, or 1MIN).
        count : int, optional
            The number of candles to fetch (maximum of 100). The default is None.
        start_time : datetime, optional
            The data start time. The default is None.
        end_time : datetime, optional
            The data end time. The default is None.
        Returns
        -------
        data : DataFrame
            The price data, as an OHLCV DataFrame.
        """

        # Check granularity was provided
        if granularity is None:
            raise Exception("Please specify candlestick granularity.")

        gran_str = granularity
        gran_map = {
            60: "1MIN",
            300: "5MINS",
            900: "15MINS",
            1800: "30MINS",
            3600: "1HOUR",
            14400: "4HOURS",
            86400: "1DAY",
        }
        granularity = gran_map[pd.Timedelta(granularity).total_seconds()]

        def fetch_between_dates():
            # Fetches data between two dates
            data = []
            start = start_time
            last = start
            timestep = pd.Timedelta(gran_str)
            while last < end_time:
                raw_data = self.api.public.get_candles(
                    instrument,
                    resolution=granularity,
                    to_iso=start.isoformat(),
                ).data["candles"]
                # Append data
                data += raw_data[::-1]

                # Increment end time
                last = datetime.strptime(data[-1]["updatedAt"], "%Y-%m-%dT%H:%M:%S.%fZ")
                start = last + 100 * timestep

                if len(raw_data) > 0:
                    # Sleep to prevent API rate-limiting
                    time.sleep(0.1)
                else:
                    start = end_time

            return data

        def fetch_count(N):
            count = min(100, N)
            raw_data = self.api.public.get_candles(
                instrument, resolution=granularity, limit=count
            ).data["candles"]
            data = raw_data[::-1]
            first_time = datetime.strptime(
                data[0]["updatedAt"], "%Y-%m-%dT%H:%M:%S.%fZ"
            )
            while len(data) < N:
                count = min(100, N - len(data))
                raw_data = self.api.public.get_candles(
                    instrument,
                    resolution=granularity,
                    to_iso=first_time.isoformat(),
                    limit=count,
                ).data["candles"]
                # Append data
                data += raw_data[::-1]
                first_time = datetime.strptime(
                    raw_data[-1]["updatedAt"], "%Y-%m-%dT%H:%M:%S.%fZ"
                )
                time.sleep(0.1)

            return data

        if count is not None:
            if start_time is None and end_time is None:
                # Fetch N most recent candles
                raw_data = fetch_count(count + 10)
            elif start_time is not None and end_time is None:
                # Fetch N candles since start_time
                raw_data = self.api.public.get_candles(
                    market=instrument,
                    resolution=granularity,
                    from_iso=start_time.isoformat(),
                    limit=count,
                ).data["candles"][::-1]
            elif end_time is not None and start_time is None:
                raw_data = self.api.public.get_candles(
                    market=instrument,
                    resolution=granularity,
                    to_iso=end_time.isoformat(),
                    limit=count,
                ).data["candles"][::-1]
            else:
                raw_data = fetch_between_dates()

        else:
            # Count is None
            try:
                assert start_time is not None and end_time is not None
                raw_data = fetch_between_dates()

            except AssertionError:
                raise Exception(
                    "When no count is provided, both start_time "
                    + "and end_time must be provided."
                )

        # Process data
        data = pd.DataFrame(raw_data)
        data.rename(
            columns={"open": "Open", "high": "High", "low": "Low", "close": "Close"},
            inplace=True,
        )
        data.index = pd.to_datetime(data["updatedAt"], format="%Y-%m-%dT%H:%M:%S.%fZ")
        data.drop(
            ["startedAt", "updatedAt", "market", "resolution"], axis=1, inplace=True
        )
        data = data.apply(pd.to_numeric, errors="ignore")
        data.drop_duplicates(inplace=True)
        data.sort_index(inplace=True)
        return data

    def _dydx_quote_data(
        self,
        data: pd.DataFrame,
        pair: str,
        granularity: str,
        start_time: datetime,
        end_time: datetime,
        count: int = None,
    ):
        """Returns the original price data for a dYdX data feed."""
        return data

    def _dydx_orderbook(self, instrument, *args, **kwargs):
        """Returns the orderbook from dYdX."""
        response = self.api.public.get_orderbook(market=instrument)
        orderbook = response.data
        return orderbook

    def _dydx_trades(self, instrument):
        """Returns public trades from dYdX."""
        response = self.api.public.get_trades(instrument)
        dydx_trades = response.data["trades"]

        # Convert to standard form
        trades = []
        for trade in dydx_trades:
            unified_trade = {
                "direction": 1 if trade["side"] == "BUY" else -1,
                "price": float(trade["price"]),
                "size": float(trade["size"]),
                "time": datetime.strptime(trade["createdAt"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            }
            trades.append(unified_trade)

        return trades

    def _none(self, *args, **kwargs):
        """Dummy method for none 'data' feed"""
        return None

    def _none_quote_data(self, *args, **kwargs):
        return None

    def _none_orderbook(self, *args, **kwargs):
        return None



import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from autotrader.brokers.broker_utils import BrokerUtils


class Utils(BrokerUtils):
    def __init__(self, **kwargs):
        return

    def update_data_with_candle(self, data, latest_candle):
        """Appends the latest candle to the data."""
        # latest candle will be in the form of a list.
        # [time, Open, High, Low, Close]
        # Need to convert this to a df.
        # latest_candle = ['2021-06-14 02:54:30', 1.20994, 1.20994, 1.20994, 1.20994]
        candle_time = datetime.strptime(latest_candle[0], "%Y-%m-%d %H:%M:%S")

        candle = pd.DataFrame(
            {
                "Open": latest_candle[1],
                "High": latest_candle[2],
                "Low": latest_candle[3],
                "Close": latest_candle[4],
            },
            index=[candle_time],
        )
        new_data = pd.concat([data, candle])

        return new_data

    def last_period(self, current_time, granularity, current_candle=False):
        """
        Returns a datetime object corresponding to the last candle that closed,
        based on the current time and the granularity of the candle.

        Note: if the current time is 10:15:01, the instrument will have just
        begun forming the 10:15 candle. As such, the latest candle in the
        instruments price history is the 10:00 candle, as that is the last
        candle that has closed.

        """

        letter_to_unit = {"S": "second", "M": "minute", "H": "hour", "D": "day"}

        letter = granularity[0]
        if len(granularity) > 1:
            number = int(granularity[1:])
        else:
            number = 1

        current_period = getattr(current_time, letter_to_unit[letter])
        if current_candle:
            last_period = number * np.floor(current_period / number)
        else:
            last_period = number * np.floor(current_period / number) - number

        if letter == "S":
            td = timedelta(
                microseconds=current_time.microsecond,
                seconds=current_time.second - last_period,
            )
        elif letter == "M":
            td = timedelta(
                microseconds=current_time.microsecond,
                seconds=current_time.second,
                minutes=current_time.minute - last_period,
            )
        elif letter == "H":
            td = timedelta(
                microseconds=current_time.microsecond,
                seconds=current_time.second,
                minutes=current_time.minute,
                hours=current_time.hour - last_period,
            )
        elif letter == "H":
            td = timedelta(
                microseconds=current_time.microsecond,
                seconds=current_time.second,
                minutes=current_time.minute,
                hours=current_time.hour,
                days=current_time.day - last_period,
            )

        last_candle_closed = current_time - td

        return last_candle_closed

    def trade_summary(self, raw_livetrade_summary, ohlc_data, granularity):
        """Constructs trade summary dataframe from Oanda trade history."""

        # Need to create a new coloumn with ohlc_data int index
        # and reutrn new trade summary, after using last_period function
        # Will also need to round the trade summary times to the nearest granularity

        # Create ohlc_data int index
        modified_data = ohlc_data
        modified_data["date"] = modified_data.index
        modified_data = modified_data.reset_index(drop=True)
        modified_data["data_index"] = modified_data.index

        # Round trade summary times to nearest candle
        rounded_times = [
            self.last_period(dt, "M15", current_candle=True)
            for dt in pd.to_datetime(raw_livetrade_summary.Date.values, utc=True)
        ]

        # Reset trade summary and add rounded times
        raw_livetrade_summary = raw_livetrade_summary.reset_index()
        raw_livetrade_summary["Date"] = rounded_times

        # Merge to preserve data int index
        livetrade_summary = pd.merge(
            modified_data, raw_livetrade_summary, left_on="date", right_on="Date"
        )

        return livetrade_summary

    def format_watchlist(self, raw_watchlist):

        watchlist = []
        for instrument in raw_watchlist:
            if str(instrument) != "nan":
                formatted_instrument = instrument[:3] + "_" + instrument[-3:]
                watchlist.append(formatted_instrument)

        return watchlist

    def get_precision(self, instrument, *args, **kwargs):
        """Returns the precision of the instrument."""
        # TODO - implement
        return {"size": 2, "price": 5}



from __future__ import annotations
import os
import pickle
import numpy as np
import pandas as pd
from decimal import Decimal
from typing import Callable
from autotrader.autodata import AutoData
from datetime import date, datetime, timezone
from autotrader.utilities import get_data_config
from autotrader.brokers.broker import AbstractBroker
from autotrader.brokers.broker_utils import OrderBook, BrokerUtils
from autotrader.brokers.trading import Order, IsolatedPosition, Position, Trade


class Broker(AbstractBroker):
    """Autotrader Virtual Broker for simulated trading.

    Attributes
    ----------
    verbosity : int
        The verbosity of the broker.
    pending_orders : dict
        A dictionary containing pending orders.
    open_orders : dict
        A dictionary containing open orders yet to be filled.
    filled_orders : dict
        A dictionary containing filled orders.
    cancelled_orders : dict
        A dictionary containing cancelled orders.
    open_trades : dict
        A dictionary containing currently open trades (fills).
    closed_trades : dict
        A dictionary containing closed trades.
    base_currency : str
        The base currency of the account. The default is 'AUD'.
    NAV : float
        The net asset value of the account.
    equity : float
        The account equity balance.
    floating_pnl : float
        The floating PnL.
    margin_available : float
        The margin available on the account.
    leverage : int
        The account leverage.
    spread : float
        The average spread to use when opening and closing trades.
    spread_units : str
        The units of the spread (eg. 'price' or 'percentage'). The default
        is 'price'.
    hedging : bool
        Flag whethere hedging is enabled on the account. The default is False.
    margin_closeout : float
        The fraction of margin available at margin call. The default is 0.
    commission_scheme : str
        The commission scheme being used ('percentage', 'fixed_per_unit'
        or 'flat'). The default is 'percentage'.
    commission : float
        The commission value associated with the commission scheme.
    maker_commission : float
        The commission value associated with liquidity making orders.
    taker_commission : float
        The commission value associated with liquidity taking orders.
    """

    def __init__(self, broker_config: dict = None, utils: BrokerUtils = None) -> None:
        """Initialise virtual broker."""
        if broker_config is not None:
            self._verbosity = broker_config["verbosity"]
        else:
            self._verbosity = 0
        self._utils = utils

        # Orders
        self._pending_orders = {}  # {instrument: {id: Order}}
        self._open_orders = {}  # {instrument: {id: Order}}
        self._filled_orders = {}
        self._cancelled_orders = {}
        self._order_id_instrument = {}  # mapper from order_id to instrument

        # Isolated positions (formerly "trades")
        # self._open_iso_pos = {}
        # self._closed_iso_pos = {}
        self._trade_id_instrument = {}  # mapper from trade_id to instrument

        # Positions
        self._positions = {}
        self._closed_positions = {}

        # Margin call flag
        self._margin_calling = False

        # Fills (executed trades)
        self._fills = []

        # Slippage model
        self._default_slippage_model = self._zero_slippage_model
        self._slippage_models = {}

        # Account
        self._base_currency = "AUD"
        self._NAV = 0  # Net asset value
        self._equity = 0  # Account equity (balance)
        self._floating_pnl = 0
        self._margin_available = 0
        self._open_interest = 0
        self._long_exposure = 0
        self._short_exposure = 0
        self._long_unrealised_pnl = 0
        self._short_unrealised_pnl = 0

        # Evolving metrics
        self._long_realised_pnl = 0
        self._short_realised_pnl = 0

        # Margin
        self._leverage = 1  # The account leverage
        self._spread = 0  # The bid/ask spread
        self._spread_units = "price"  # The units of the spread
        self._hedging = False  # Allow simultaneous trades on opposing sides
        self._margin_closeout = 0.0  # Fraction at margin call

        # Funding rate (for perpetual contracts)
        self._charge_funding = False
        self._funding_rate_history = None
        self._funding_history = []
        self._update_freq = None  # Backtest trading update frequency

        # Commissions
        self._commission_scheme = (
            "percentage"  # Either percentage, fixed_per_unit or flat
        )
        self._commission = 0
        self._maker_commission = 0  # Liquidity 'maker' trade commission
        self._taker_commission = 0  # Liquidity 'taker' trade commission

        # History
        self._latest_time = None

        # Last order and trade counts
        self._last_order_id = 0
        self._last_trade_id = 0
        self._last_fill_id = 0

        # Paper trading mode
        self._paper_trading = False  # Paper trading mode boolean
        self._public_trade_access = False  # Use public trades to update orders
        self.autodata = None  # AutoData instance
        self._state = None  # Last state snapshot
        self._picklefile = None  # Pickle filename

        # CCXT unification
        self.exchange = ""

    def __repr__(self):
        data_feed = self.autodata._feed
        if data_feed == "ccxt":
            data_feed = self.autodata._ccxt_exchange
        return f"AutoTrader Virtual Broker ({data_feed} data feed)"

    def __str__(self):
        return self.__repr__()

    def configure(
        self,
        verbosity: int = None,
        initial_balance: float = None,
        leverage: int = None,
        spread: float = None,
        spread_units: str = None,
        commission: float = None,
        commission_scheme: str = None,
        maker_commission: float = None,
        taker_commission: float = None,
        hedging: bool = None,
        base_currency: str = None,
        paper_mode: bool = None,
        public_trade_access: bool = None,
        margin_closeout: float = None,
        default_slippage_model: Callable = None,
        slippage_models: dict = None,
        charge_funding: bool = None,
        funding_history: pd.DataFrame = None,
        autodata_config: dict = None,
        picklefile: str = None,
        **kwargs,
    ):
        """Configures the broker and account settings.

        Parameters
        ----------
        verbosity : int, optional
            The verbosity of the broker. The default is 0.
        initial_balance : float, optional
            The initial balance of the account, specified in the
            base currency. The default is 0.
        leverage : int, optional
            The leverage available. The default is 1.
        spread : float, optional
            The bid/ask spread to use in backtest (specified in units
            defined by the spread_units argument). The default is 0.
        spread_units : str, optional
            The unit of the spread specified. Options are 'price', meaning
            that the spread is quoted in price units, or 'percentage',
            meaning that the spread is quoted as a percentage of the
            market price. The default is 'price'.
        commission : float, optional
            Trading commission as percentage per trade. The default is 0.
        commission_scheme : str, optional
            The method with which to apply commissions to trades made. The
            options are (1) 'percentage', where the percentage specified by
            the commission argument is applied to the notional trade value,
            (2) 'fixed_per_unit', where the monetary value specified by the
            commission argument is multiplied by the number of units in the
            trade, and (3) 'flat', where a flat monetary value specified by
            the commission argument is charged per trade made, regardless
            of size. The default is 'percentage'.
        maker_commission : float, optional
            The commission to charge on liquidity-making orders. The default
            is None, in which case the nominal commission argument will be
            used.
        taker_commission: float, optional
            The commission to charge on liquidity-taking orders. The default
            is None, in which case the nominal commission argument will be
            used.
        hedging : bool, optional
            Allow hedging in the virtual broker (opening simultaneous
            trades in oposing directions). The default is False.
        base_currency : str, optional
            The base currency of the account. The default is AUD.
        paper_mode : bool, optional
            A boolean flag to indicate if the broker is in paper trade mode.
            The default is False.
        public_trade_access : bool, optional
            A boolean flag to signal if public trades are being used to
            update limit orders. The default is False.
        margin_closeout : float, optional
            The fraction of margin usage at which a margin call will occur.
            The default is 0.
        default_slippage_model : Callable, optional
            The default model to use when calculating the percentage slippage
            on the fill price, for a given order size. The default functon
            returns zero.
        slippage_models : dict, optional
            A dictionary of callable slippage models, keyed by instrument.
        charge_funding : bool, optional
            A boolean flag to charge funding rates. The default is False.
        funding_history : pd.DataFrame, optional
            A DataFrame of funding rate histories for instruments being traded,
            to backtest trading perpetual futures.
            This is a single frame with as many columns as instruments being
            traded. If an instrument is not present, the funding rate will be
            zero.
        picklefile : str, optional
            The filename of the picklefile to load state from. If you do not
            wish to load from state, leave this as None. The default is None.
        """
        self._verbosity = verbosity if verbosity is not None else self._verbosity
        self._leverage = leverage if leverage is not None else self._leverage
        self._commission = commission if commission is not None else self._commission
        self._commission_scheme = (
            commission_scheme
            if commission_scheme is not None
            else self._commission_scheme
        )
        self._spread = spread if spread is not None else self._spread
        self._spread_units = (
            spread_units if spread_units is not None else self._spread_units
        )
        self._base_currency = (
            base_currency if base_currency is not None else self._base_currency
        )
        self._paper_trading = (
            paper_mode if paper_mode is not None else self._paper_trading
        )
        self._public_trade_access = (
            public_trade_access
            if public_trade_access is not None
            else self._public_trade_access
        )
        self._margin_closeout = (
            margin_closeout if margin_closeout is not None else self._margin_closeout
        )
        self._hedging = hedging if hedging is not None else self._hedging
        self._picklefile = picklefile if picklefile is not None else self._picklefile

        # Assign commissions for making and taking liquidity
        self._maker_commission = (
            maker_commission if maker_commission is not None else self._commission
        )
        self._taker_commission = (
            taker_commission if taker_commission is not None else self._commission
        )

        # Configure slippage models
        self._default_slippage_model = (
            default_slippage_model
            if default_slippage_model is not None
            else self._default_slippage_model
        )
        self._slippage_models = (
            slippage_models if slippage_models is not None else self._slippage_models
        )

        # Configure funding rate mechanics
        self._charge_funding = (
            charge_funding if charge_funding is not None else self._charge_funding
        )
        self._funding_rate_history = (
            funding_history
            if funding_history is not None
            else self._funding_rate_history
        )

        if autodata_config is not None:
            # Instantiate AutoData from config
            data_config = get_data_config(
                **autodata_config,
            )
            self.autodata = AutoData(data_config, **autodata_config)

        else:
            # Create local data instance
            self.autodata = AutoData()

        # Initialise balance
        if initial_balance is not None:
            self._make_deposit(initial_balance)

        # Check for pickled state
        if self._paper_trading and self._picklefile is not None:
            # Load state
            if os.path.exists(picklefile):
                self._load_state()

    def get_NAV(self) -> float:
        """Returns Net Asset Value of account."""
        return self._NAV

    def get_balance(self) -> float:
        """Returns balance of account."""
        return self._equity

    def place_order(self, order: Order, **kwargs) -> None:
        """Place order with broker."""
        # Call order to set order time
        datetime_stamp = (
            kwargs["order_time"]
            if "order_time" in kwargs
            else datetime.now(timezone.utc)
        )
        order(order_time=datetime_stamp)

        # Define reference price
        if order.order_type == "limit" or order.order_type == "stop-limit":
            # Use limit price
            ref_price = order.order_limit_price

        else:
            # Use order price
            ref_price = order.order_price

        # Convert stop distance to price
        if order.stop_loss is None and order.stop_distance:
            order.stop_loss = (
                ref_price - order.direction * order.stop_distance * order.pip_value
            )

        # Verify SL price
        invalid_order = False
        if order.stop_loss and order.direction * (ref_price - order.stop_loss) < 0:
            direction = "long" if order.direction > 0 else "short"
            SL_placement = "below" if order.direction > 0 else "above"
            reason = (
                "Invalid stop loss request: stop loss must be "
                + f"{SL_placement} the order price for a {direction}"
                + " trade order.\n"
                + f"Order Price: {ref_price}\nStop Loss: {order.stop_loss}"
            )
            invalid_order = True

        # Verify TP price
        if (
            order.take_profit is not None
            and order.direction * (ref_price - order.take_profit) > 0
        ):
            direction = "long" if order.direction > 0 else "short"
            TP_placement = "above" if order.direction > 0 else "below"
            reason = (
                "Invalid take profit request: take profit must be "
                + f"{TP_placement} the order price for a {direction}"
                + " trade order.\n"
                + f"Order Price: {ref_price}\nTake Profit: {order.take_profit}"
            )
            invalid_order = True

        # Verify order size
        if order.order_type in ["market", "limit", "stop-limit"] and order.size == 0:
            # Invalid order size
            reason = "Invalid order size (must be non-zero)."
            invalid_order = True

        # Check limit order does not cross book
        try:
            if order.order_type in ["limit"]:
                if self._paper_trading:
                    # Get live midprice
                    orderbook = self.get_orderbook(order.instrument)
                    ref_price = (
                        float(orderbook["bids"][0]["price"])
                        + float(orderbook["asks"][0]["price"])
                    ) / 2
                else:
                    # Use order / stop price
                    ref_price = (
                        order.order_stop_price
                        if order.order_stop_price is not None
                        else order.order_price
                    )
                invalid_order = (
                    order.direction * (ref_price - order.order_limit_price) < 0
                )
                reason = (
                    f"Invalid limit price for {order.__repr__()} "
                    + f"(reference price: {ref_price}, "
                    + f"limit price: {order.order_limit_price})"
                )
        except:
            # Exception, continue
            pass

        # Assign order ID
        order.id = self._get_new_order_id()
        self._order_id_instrument[order.id] = order.instrument

        # Add order to pending_orders dict
        order.status = "pending"
        try:
            self._pending_orders[order.instrument][order.id] = order
        except KeyError:
            # Instrument hasn't been in pending orders yet
            self._pending_orders[order.instrument] = {order.id: order}

        # Submit order
        if invalid_order:
            # Invalid order, cancel it
            self.cancel_order(order.id, reason, "_pending_orders", datetime_stamp)

        else:
            # Move order to open_orders or leave in pending
            immediate_orders = []
            if order.order_type in immediate_orders or self._paper_trading:
                # Move to open orders
                self._move_order(
                    order,
                    from_dict="_pending_orders",
                    to_dict="_open_orders",
                    new_status="open",
                )

            # Print
            if self._verbosity > 0:
                print(
                    f"{datetime_stamp}: Order {order.id} received: {order.__repr__()}"
                )

    def get_orders(self, instrument: str = None, order_status: str = "open") -> dict:
        """Returns orders of status order_status."""
        all_orders = getattr(self, f"_{order_status}_orders")
        if instrument:
            # Return orders for instrument specified
            try:
                orders = all_orders[instrument]
            except KeyError:
                # There are currently no orders for this instrument
                orders = {}
        else:
            # Return all orders
            orders = {}
            for instr, instr_orders in all_orders.items():
                orders.update(instr_orders)
        return orders.copy()

    def cancel_order(
        self,
        order_id: int,
        reason: str = None,
        from_dict: str = "_open_orders",
        timestamp: datetime = None,
        **kwargs,
    ) -> None:
        """Cancels the order.

        Parameters
        ----------
        order_id : int
            The ID of the order to be cancelled.
        reason : str, optional
            The reason why the order is being cancelled. The default
            is None.
        from_dict: str, optional
            The dictionary currently holding the order. The default is
            'open_orders'.
        timestamp: datetime, optional
            The datetime stamp of the order cancellation. The default is
            None.
        """

        instrument = self._order_id_instrument[order_id]
        from_dict = getattr(self, from_dict)[instrument]
        reason = reason if reason is not None else "User cancelled."

        if instrument not in self._cancelled_orders:
            # Initialise instrument in cancelled_orders
            self._cancelled_orders[instrument] = {}
        self._cancelled_orders[instrument][order_id] = from_dict.pop(order_id)
        self._cancelled_orders[instrument][order_id].reason = reason
        self._cancelled_orders[instrument][order_id].status = "cancelled"

        if self._verbosity > 0 and reason:
            # Print cancel reason to console
            print(f"{timestamp}: Order {order_id} cancelled: {reason}")

    def get_trades(self, instrument: str = None, **kwargs) -> dict:
        """Returns fills for the specified instrument.

        Parameters
        ----------
        instrument : str, optional
            The instrument to fetch trades under. The default is None.
        """

        # Get all instrument fills
        all_trades = self._fills

        if instrument:
            # Specific instrument(s) requested
            trades = [fill for fill in all_trades if fill.instrument == "EUR"]

        else:
            # Return all trades
            trades = all_trades

        # Convert to a dict with keys by id
        trades_dict = {fill.id: fill for fill in all_trades}

        # Return a copy to prevent unintended manipulation
        return trades_dict.copy()

    def get_positions(self, instrument: str = None) -> dict:
        """Returns the positions held by the account, sorted by
        instrument.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument name (symbol). If 'None' is provided,
            all positions will be returned. The default is None.

        Returns
        -------
        open_positions : dict
            A dictionary containing details of the open positions.

        Notes
        ------
        net_position: refers to the number of units held in the position.

        """
        if instrument:
            # Instrument provided
            if instrument in self._positions:
                return {instrument: self._positions[instrument]}
            else:
                return {}
        else:
            return self._positions.copy()

    def get_margin_available(self) -> float:
        """Returns the margin available on the account."""
        return self._margin_available

    def get_orderbook(self, instrument: str, midprice: float = None) -> OrderBook:
        """Returns the orderbook."""
        # Get public orderbook
        if self._paper_trading:
            # Papertrading, try get realtime orderbook
            try:
                orderbook = self.autodata.L2(
                    instrument, spread_units=self._spread_units, spread=self._spread
                )
            except Exception as e:
                # Fetching orderbook failed, revert to local orderbook
                print("Exception:", e)
                print("  Instrument:", instrument)
                orderbook = self.autodata._local_orderbook(
                    instrument=instrument,
                    spread_units=self._spread_units,
                    spread=self._spread,
                    midprice=midprice,
                )

        else:
            # Backtesting, use local pseudo-orderbook
            orderbook = self.autodata._local_orderbook(
                instrument=instrument,
                spread_units=self._spread_units,
                spread=self._spread,
                midprice=midprice,
            )

        # TODO - Add local orders to the book?

        return orderbook

    def _update_positions(
        self,
        instrument: str,
        candle: pd.Series = None,
        L1: dict = None,
        trade: dict = None,
    ) -> None:
        """Updates orders and open positions based on the latest data.

        Parameters
        ----------
        instrument : str
            The name of the instrument being updated.
        candle : pd.Series
            An OHLC candle used to update orders and trades.
        L1 : dict, optional
            A dictionary a containing level 1 price snapshot to update
            the positions with. This dictionary must have the keys
            'bid', 'ask', 'bid_size' and 'ask_size'.
        trade : dict, optional
            A public trade, used to update virtual limit orders.
        """

        def stop_trigger_condition(order_stop_price, order_direction):
            """Returns True if the order stop price has been triggered
            else False."""
            if L1 is not None:
                # Use L1 data to trigger
                reference_price = L1["bid"] if order_direction > 0 else L1["ask"]
                triggered = order_direction * (reference_price - order_stop_price) > 0

            else:
                # Use OHLC data to trigger
                triggered = candle.Low < order_stop_price < candle.High
            return triggered

        def get_last_price(trade_direction):
            """Returns the last reference price for a trade. If the
            trade is long, this will refer to the bid price. If short,
            this refers to the ask price."""
            if L1 is not None:
                last_price = L1["bid"] if trade_direction > 0 else L1["ask"]
            else:
                last_price = candle.Close
            return last_price

        def get_market_ref_price(order_direction):
            """Returns the reference price for a market order."""
            if L1 is not None:
                reference_price = L1["bid"] if order_direction > 0 else L1["ask"]
            else:
                reference_price = candle.Open
            return reference_price

        def limit_trigger_condition(order_direction, order_limit_price):
            """Returns True if the order limit price has been triggered
            else False."""
            if L1 is not None:
                # Use L1 data to trigger (based on midprice)
                ref_price = (L1["bid"] + L1["ask"]) / 2
            else:
                # Use OHLC data to trigger
                ref_price = candle.Low if order_direction > 0 else candle.High
            triggered = order_direction * (ref_price - float(order_limit_price)) <= 0
            return triggered

        def process_orders_in_dict(orders):
            for order_id, order in orders.items():
                # Check if order has been cancelled (OCO)
                if order.id not in self._open_orders[order.instrument]:
                    continue

                # Order is still open, process it
                if order.order_type == "market":
                    # Market order type - proceed to fill
                    reference_price = get_market_ref_price(order.direction)
                    if order.order_price is None:
                        order.order_price = reference_price
                    self._process_order(
                        order=order,
                        fill_time=latest_time,
                        reference_price=reference_price,
                    )

                elif "stop" in order.order_type:
                    # Check if order_stop_price has been reached yet
                    if stop_trigger_condition(order.order_stop_price, order.direction):
                        # order_stop_price has been reached
                        if order.order_type == "stop-limit":
                            # Change order type to 'limit'
                            order.order_type = "limit"
                        else:
                            # Stop order triggered - proceed to market fill
                            reference_price = order.order_stop_price
                            order.order_price = reference_price
                            order.order_type = "market"
                            self._process_order(
                                order=order,
                                fill_time=latest_time,
                                reference_price=reference_price,
                            )

                # Check limit orders
                if order.order_type == "limit":
                    # Limit order type
                    if not self._public_trade_access:
                        # Update limit orders based on price feed
                        triggered = limit_trigger_condition(
                            order.direction, order.order_limit_price
                        )
                        if triggered:
                            # Limit price triggered, proceed
                            self._process_order(
                                order=order,
                                fill_time=latest_time,
                                reference_price=order.order_limit_price,
                            )
                    else:
                        # Update limit orders based on trade feed
                        if trade is not None:
                            self._public_trade(instrument, trade)

        # Check for data availability
        if L1 is not None:
            # Using L1 data to update
            latest_time = datetime.now(timezone.utc)

        elif candle is not None:
            # Using OHLC data to update
            latest_time = candle.name

        else:
            # No new price data
            if trade is None:
                # No trade either, exit
                return
            else:
                # Public trade received, only update limit orders
                self._public_trade(instrument, trade)
                return

        # Open pending orders
        pending_orders = self.get_orders(instrument, "pending")
        for order_id, order in pending_orders.items():
            if latest_time > order.order_time:
                self._move_order(
                    order,
                    from_dict="_pending_orders",
                    to_dict="_open_orders",
                    new_status="open",
                )

        # Update open orders for current instrument
        open_orders = self.get_orders(instrument)
        process_orders_in_dict(open_orders)

        # Check for any SL/TP orders spawned
        currently_open_orders = self.get_orders(instrument=instrument)
        newly_opened_orders = dict(
            set(currently_open_orders.items()) - set(open_orders.items())
        )
        process_orders_in_dict(newly_opened_orders)

        # Update position
        position = self.get_positions(instrument=instrument)
        if position:
            position = position[instrument]
            # Position held, update it
            position.last_price = get_last_price(np.sign(position.net_position))
            position.last_time = latest_time
            position.notional = position.last_price * abs(position.net_position)
            position.pnl = (
                position.net_position
                * (position.last_price - position.avg_price)
                * position.HCF
            )

            # Charge funding fees
            # if self._charge_funding:
            #     # TODO - move to unified backtest/livetrade method
            #     last_settlement = self._funding_rate_history.index.get_indexer([pd.Timestamp(latest_time).floor(self._update_freq)], method="nearest")[0]
            #     last_rate = self._funding_rate_history.iloc[last_settlement][instrument]

            #     # Set to zero to prevent re-charging
            #     self._funding_rate_history.iloc[last_settlement][instrument] = 0

        # Update floating pnl and margin available
        self._update_margin(
            instrument=instrument,
            latest_time=latest_time,
        )

        # Update open position value
        self._NAV = self._equity + self._floating_pnl

        # Update account history
        self._latest_time = latest_time

        # Save state
        if self._paper_trading and self._picklefile is not None:
            self._save_state()

    def _update_all(self):
        """Convenience method to update all open positions when paper trading."""
        # Update orders
        for instrument in self._open_orders:
            l1 = self.autodata.L1(instrument=instrument)
            self._update_positions(instrument=instrument, L1=l1)

        # Update positions
        for instrument in self._positions.keys():
            l1 = self.autodata.L1(instrument=instrument)
            self._update_positions(instrument=instrument, L1=l1)

    def _update_instrument(self, instrument):
        """Convenience method to update a single instrument when paper
        trading.
        """
        # Check for existing orders or position in this instrument
        orders = self.get_orders(instrument=instrument)
        position = self.get_positions(instrument=instrument)
        if len(orders) + len(position) > 0:
            # Order or position exists for this instrument, update it
            l1 = self.autodata.L1(instrument=instrument)
            self._update_positions(instrument=instrument, L1=l1)

    def _process_order(
        self,
        order: Order,
        fill_time: datetime = None,
        reference_price: float = None,
        trade_size: float = None,
    ):
        """Processes an order, either filling or cancelling it.

        Parameters
        -----------
        order : Order
            The order being processed.
        fill_time : datetime, optional
            The time to fill the order.
        reference_price : float, optional
            The order reference price (either market price or order limit price).
        trade_size : float, optional
            The size of a public trade being used to fill orders (papertrade
            mode). The default is None.
        """

        # Check if papertrading
        if self._paper_trading:
            # Use current time as fill time
            tz = fill_time.tzinfo  # inherit timezone of dataset
            if tz is not None:
                fill_time = datetime.now(tz=tz)

        # Check for public trade to fill order
        if trade_size is not None:
            # Fill limit order with trade_size provided
            if trade_size < order.size:
                # Create a new order for the portion to be filled by the trade
                order = Order._partial_fill(order=order, units_filled=trade_size)

                # Assign new order ID
                order.id = self._get_new_order_id()
                self._order_id_instrument[order.id] = order.instrument

                # Move new order to open_orders
                # The original order will remain with reduced size
                self._open_orders[order.instrument][order.id] = order

        order_notional = order.size * reference_price * order.HCF
        margin_required = self._calculate_margin(order_notional)

        if margin_required < self._margin_available or order.reduce_only:
            # Order can be filled
            if order.order_type == "limit":
                # Limit order, use limit price for execution
                fill_price = order.order_limit_price

            elif order.order_type == "market":
                # Market order, trade through the book using reference size
                fill_price = self._trade_through_book(
                    instrument=order.instrument,
                    direction=order.direction,
                    size=order.size,
                    reference_price=reference_price,
                    precision=order.price_precision,
                )
            else:
                # Unrecognised order type
                raise Exception(f"Unrecognised order type: {order.order_type}")

            # Fill order
            self._fill_order(
                last_price=reference_price,
                order=order,
                fill_price=fill_price,
                fill_time=fill_time,
            )

        else:
            # Cancel order
            cancel_reason = (
                "Insufficient margin to fill order "
                + f"(${margin_required} required, ${self._margin_available} "
                + "available)."
            )
            self.cancel_order(
                order_id=order.id, reason=cancel_reason, timestamp=fill_time
            )

    def _modify_position(self, trade: Trade, reduce_only: bool):
        """Modifies the position in a position."""
        if trade.instrument in self._positions:
            # Instrument already has a position
            price_precision = self._positions[trade.instrument].price_precision
            starting_net_position = self._positions[trade.instrument].net_position

            if reduce_only:
                # Make sure the trade can only reduce (not swap) the position
                trade.size = min(
                    trade.size,
                    abs(self._positions[trade.instrument].net_position),
                )

                if trade.direction == np.sign(starting_net_position):
                    # The reduce order is on the wrong side, cancel the fill
                    del self._fills[-1]
                    return

            # Update the position with the trade
            self._positions[trade.instrument]._update_with_fill(
                trade=trade,
            )
            new_net_position = self._positions[trade.instrument].net_position

            # Check if position has been reduced
            if np.sign(new_net_position - starting_net_position) != np.sign(
                starting_net_position
            ):
                # Update account with position pnl
                units_reduced = min(
                    abs(new_net_position - starting_net_position),
                    abs(starting_net_position),
                )
                reduction_direction = np.sign(starting_net_position - new_net_position)
                pnl = (
                    reduction_direction
                    * units_reduced
                    * (
                        trade.fill_price
                        - self._positions[trade.instrument]._prev_avg_price
                    )
                )
                self._adjust_balance(pnl)

                # Update realised PnL metrics
                if np.sign(starting_net_position) > 0:
                    # Long position reduced
                    self._long_realised_pnl += pnl
                else:
                    # Short position reduced
                    self._short_realised_pnl += pnl

            # Check if position is zero
            if round(new_net_position, price_precision) == 0:
                # Move to closed positions (and add exit time)
                popped_position = self._positions.pop(trade.instrument)
                popped_position.exit_time = trade.fill_time
                popped_position.exit_price = trade.fill_price
                if trade.instrument in self._closed_positions:
                    # Append
                    self._closed_positions[trade.instrument].append(popped_position)
                else:
                    # Create new entry
                    self._closed_positions[trade.instrument] = [popped_position]

            elif np.sign(self._positions[trade.instrument].net_position) != np.sign(
                starting_net_position
            ):
                # Position has swapped sides
                pass

        else:
            # Create new position
            self._positions[trade.instrument] = Position._from_fill(
                trade=trade,
            )

    def _fill_order(
        self,
        last_price: float,
        fill_price: float,
        fill_time: datetime,
        order: Order = None,
        liquidation_order: bool = False,
    ) -> None:
        """Marks an order as filled and records the trade as a Fill.

        Parameters
        ----------
        fill_price : float
            The fill price.
        fill_time : datetime
            The time at which the order is filled.
        order : Order, optional
            The order to fill. The default is None, in which case the arguments
            below must be specified.
        liquidation_order : bool, optional
            A flag whether this is a liquidation order from the broker.
        """
        if not liquidation_order:
            # Filling an order changes its status to 'filled'
            self._move_order(
                order=order,
                from_dict="_open_orders",
                to_dict="_filled_orders",
                new_status="filled",
            )

        # Infer attributes from provided order
        instrument = order.instrument
        order_price = order.order_price
        order_time = order.order_time
        order_size = order.size
        order_type = order.order_type
        direction = order.direction
        order_id = order.id
        HCF = order.HCF
        fill_price = round(fill_price, order.price_precision)

        # Check for SL
        if order.stop_loss is not None:
            # Create SL order
            sl_order = Order(
                instrument=order.instrument,
                direction=-order.direction,
                size=abs(order.size),
                order_type="stop",
                order_stop_price=round(order.stop_loss, order.price_precision),
                reduce_only=True,
                id=self._get_new_order_id(),
                parent_order=order.id,
                status="open",
                order_time=fill_time,
                order_price=fill_price,
            )
            # Add to open orders
            try:
                self._open_orders[sl_order.instrument][sl_order.id] = sl_order
            except KeyError:
                self._open_orders[sl_order.instrument] = {sl_order.id: sl_order}

            # Add to map
            self._order_id_instrument[sl_order.id] = sl_order.instrument

        # Check for TP
        if order.take_profit is not None:
            # Create TP order
            tp_order = Order(
                instrument=order.instrument,
                direction=-order.direction,
                size=abs(order.size),
                order_type="limit",
                order_limit_price=round(order.take_profit, order.price_precision),
                reduce_only=True,
                id=self._get_new_order_id(),
                parent_order=order.id,
                status="open",
                price_precision=order.price_precision,
                size_precision=order.size_precision,
                order_time=fill_time,
                order_price=fill_price,
            )
            # Add to open orders
            try:
                self._open_orders[tp_order.instrument][tp_order.id] = tp_order
            except KeyError:
                self._open_orders[tp_order.instrument] = {tp_order.id: tp_order}

            # Add to map
            self._order_id_instrument[tp_order.id] = tp_order.instrument

        # Link SL to TP OCO style
        if order.stop_loss is not None and order.take_profit is not None:
            tp_order.OCO.append(sl_order.id)
            sl_order.OCO.append(tp_order.id)

        # Charge trading fees
        commission = self._calculate_commissions(
            price=fill_price, units=order_size, HCF=HCF, order_type=order_type
        )
        self._adjust_balance(-commission, latest_time=fill_time)

        # Create Trade and append to fills
        # Note that this object may be modified by _modify_position
        trade = Trade(
            instrument=instrument,
            order_price=order_price,
            order_time=order_time,
            size=order_size,
            last_price=last_price,
            fill_time=fill_time,
            fill_price=fill_price,
            fill_direction=direction,
            fee=commission,
            id=self._get_new_fill_id(),
            order_id=order_id,
            order_type=order_type,
            _price_precision=order.price_precision,
            _size_precision=order.size_precision,
        )
        self._fills.append(trade)

        # Check OCO orders
        for order_id in order.OCO:
            self.cancel_order(
                order_id=order_id,
                reason="Linked order cancellation.",
                timestamp=fill_time,
            )

        # Print fill to console
        if self._verbosity > 0:
            id_str = order_id if order_id is not None else ""
            side = "(Buy)" if direction > 0 else "(Sell)"
            fill_str = (
                f"{fill_time}: Order {id_str} filled: {order_size} "
                + f"units of {instrument} @ {fill_price} {side}"
            )
            print(fill_str)

        # Update position with fill
        self._modify_position(trade=trade, reduce_only=order.reduce_only)

    def _move_order(
        self,
        order: Order,
        from_dict: str = "_open_orders",
        to_dict: str = "_filled_orders",
        new_status: str = "filled",
    ) -> None:
        """Moves an order from the from_dict to the to_dict."""
        order.status = new_status
        from_dict = getattr(self, from_dict)[order.instrument]
        to_dict = getattr(self, to_dict)
        popped_item = from_dict.pop(order.id)
        try:
            to_dict[order.instrument][order.id] = popped_item
        except KeyError:
            to_dict[order.instrument] = {order.id: popped_item}

    def _trade_through_book(
        self,
        instrument: str,
        direction: int,
        size: float,
        reference_price: float = None,
        precision: int = None,
    ) -> float:
        """Returns an average fill price by filling an order through
        the orderbook.

        Parameters
        -----------
        instrument : str
            The instrument to fetch the orderbook for.
        direction : int
            The direction of the trade (1 for long, -1 for short). Used
            to specify either bid or ask prices.
        size : float
            The size of the trade.
        reference_price : float, optional
            The reference price to use if artificially creating an
            orderbook.
        precision : dict, optional
            The precision to use for rounding prices. The default
            is None.
        """
        # Get order book
        book = self.get_orderbook(instrument, reference_price)

        # Work through the order book
        units_to_fill = size
        side = "bids" if direction < 0 else "asks"
        fill_prices = []
        fill_sizes = []
        level_no = 0
        while units_to_fill > 0:
            # Consume liquidity
            level = getattr(book, side).iloc[level_no]
            units_consumed = min(units_to_fill, float(level["size"]))
            fill_prices.append(float(level["price"]))
            fill_sizes.append(units_consumed)

            # Iterate
            level_no += 1
            units_to_fill -= units_consumed

        avg_fill_price = sum(
            [fill_sizes[i] * fill_prices[i] for i in range(len(fill_prices))]
        ) / sum(fill_sizes)

        # Apply slippage function
        if not self._paper_trading:
            # Currently backtesting - apply slippage function to fill price
            slippage_model = self._slippage_models.setdefault(
                instrument,
                self._default_slippage_model,
            )
            slippage_pc = slippage_model(size)
            avg_fill_price *= 1 + direction * slippage_pc

        if precision is not None:
            # Round price to precision
            avg_fill_price = round(avg_fill_price, precision)

        return avg_fill_price

    def _calculate_commissions(
        self,
        price: float,
        units: float = None,
        HCF: float = 1,
        order_type: str = "market",
    ) -> float:
        """Calculates trade commissions."""
        # Get appropriate commission value
        commission_val = (
            self._taker_commission if order_type == "market" else self._maker_commission
        )

        if self._commission_scheme == "percentage":
            # Commission charged as percentage of trade value
            trade_value = abs(units) * float(price) * HCF
            commission = (commission_val / 100) * trade_value

        elif self._commission_scheme == "fixed_per_unit":
            # Fixed commission per unit traded
            commission = commission_val * units

        elif self._commission_scheme == "flat":
            # Flat commission value per trade
            commission = commission_val

        return commission

    def _adjust_balance(self, amount: float, latest_time: datetime = None) -> None:
        """Adjusts the balance of the account."""
        self._equity += amount
        self._update_margin(latest_time=latest_time)

    def _make_deposit(self, deposit: float) -> None:
        """Adds deposit to account balance and NAV."""
        self._equity += deposit
        self._NAV += deposit
        self._update_margin()

    def _calculate_margin(self, position_value: float) -> float:
        """Calculates margin required to take a position with the
        available leverage of the account.
        """
        margin = position_value / self._leverage
        return margin

    def _update_margin(
        self, instrument: str = None, latest_time: datetime = None
    ) -> None:
        """Updates the margin available in the account."""
        # TODO - only update with instrument
        margin_used = 0
        floating_pnl = 0
        open_interest = 0

        positions = self.get_positions()
        long_exposure = 0
        short_exposure = 0
        long_upnl = 0
        short_upnl = 0
        for instrument, position in positions.items():
            margin_used += self._calculate_margin(position.notional)
            floating_pnl += position.pnl
            open_interest += position.notional

            if position.direction > 0:
                long_upnl += position.pnl
                long_exposure += position.notional
            else:
                short_upnl += position.pnl
                short_exposure += position.notional

        # Update unrealised PnL
        self._long_unrealised_pnl = long_upnl
        self._short_unrealised_pnl = short_upnl
        self._long_exposure = long_exposure
        self._short_exposure = short_exposure
        self._floating_pnl = floating_pnl

        # Update margin available
        self._margin_available = self._NAV - margin_used

        # Update open interest
        self._open_interest = open_interest

        # Check for margin call
        if (
            self._leverage > 1
            and self._margin_available / self._NAV < self._margin_closeout
            and not self._margin_calling
        ):
            # Margin call - close all positions
            self._margin_calling = True
            if self._verbosity > 0:
                print("MARGIN CALL: closing all positions.")
            positions = self.get_positions()
            for instrument, position in positions.items():
                last_price = position.last_price
                self._margin_call(instrument, latest_time, last_price)

            # Reset margin call flag
            self._margin_calling = False

    def _get_new_order_id(self):
        self._last_order_id += 1
        return self._last_order_id

    def _get_new_trade_id(self):
        self._last_trade_id += 1
        return self._last_trade_id

    def _get_new_fill_id(self):
        self._last_fill_id += 1
        return self._last_fill_id

    def _margin_call(self, instrument: str, latest_time: datetime, latest_price: float):
        """Closes the open position of the instrument."""
        # Construct full position closeout market order
        position = self.get_positions(instrument=instrument)[instrument]
        size = abs(position.net_position)
        direction = np.sign(position.net_position)
        ref_time = latest_time
        ref_price = latest_price
        ref_id = self._get_new_order_id()
        closeout_order = Order(
            instrument=instrument,
            order_price=ref_price,
            order_time=ref_time,
            direction=-direction,
            size=size,
            order_type="market",
            id=ref_id,
            HCF=1,
        )

        # Fire the order
        fill_price = self._trade_through_book(
            instrument=instrument,
            direction=-direction,
            size=size,
            reference_price=latest_price,
        )
        self._fill_order(
            last_price=latest_price,
            fill_price=fill_price,
            fill_time=ref_time,
            order=closeout_order,
            liquidation_order=True,
        )

    def _add_orders_to_book(self, instrument, orderbook):
        """Adds local orders to the orderbook."""
        # TODO - implement
        orders = self.get_orders(instrument)
        for order in orders:
            if order.order_type == "limit":
                side = "bids" if order.direction > 0 else "asks"

                # Add to the book
                orderbook[side]

        return orderbook

    @staticmethod
    def _zero_slippage_model(*args, **kwargs):
        "Returns zero slippage."
        return 0

    def _save_state(self):
        """Pickles the current state of the broker."""
        try:
            # Remove old picklefile (if it exists)
            os.remove(self._picklefile)
        except:
            pass

        with open(self._picklefile, "wb") as file:
            pickle.dump(self, file)

    def _load_state(self):
        """Loads the state of the broker from a pickle."""
        verbosity = self._verbosity
        try:
            with open(self._picklefile, "rb") as file:
                state = pickle.load(file)

            # Overwrite present state from pickle
            for key, item in state.__dict__.items():
                self.__dict__[key] = item

            if verbosity > 0:
                print("Virtual broker state loaded from pickle.")
        except:
            print("Failed to load virtual broker state.")

    def _public_trade(self, instrument: str, trade: dict):
        """Uses a public trade to update virtual orders."""
        # TODO - use a Trade object
        trade_direction = trade["direction"]
        trade_price = trade["price"]
        trade_size = trade["size"]
        trade_time = trade["time"]

        trade_units_remaining = trade_size
        open_orders = self.get_orders(instrument).copy()
        for order_id, order in open_orders.items():
            if order.order_type == "limit":
                if order.direction != trade_direction:
                    # Buy trade for sell orders, Sell trade for buy orders
                    order_price = Decimal(str(order.order_limit_price)).quantize(
                        Decimal(str(trade_price))
                    )
                    if trade_price == order_price and trade_units_remaining > 0:
                        # Fill as much as possible
                        trade_units_consumed = min(trade_units_remaining, order.size)
                        self._process_order(
                            order=order,
                            fill_time=trade_time,
                            reference_price=order.order_limit_price,
                            trade_size=trade_units_consumed,
                        )

                        # Update trade_units_remaining
                        trade_units_remaining -= trade_units_consumed
